
"""
Optimisation of the parameters for the Zumbach test

From:
    Zumbach, G., (2007).
    Time reversal invariance in finance.
    https://arxiv.org/pdf/0708.4022.pdf
"""

import numpy as np
from itertools import product
from multiprocessing import Pool
from functools import partial

from irreversibility.Metrics.Zumbach import GetPValue


def Optimisation( tsSet, paramSet = None, criterion = np.median, numProcesses = -1, **kwargs ):

    """
    Optimise the parameters for the Zumbach test.

    Parameters
    ----------
    tsSet : list of numpy.array
        Time series to be used in the optimisation.
    paramSet : list
        List of parameters' values to be evaluated. For each parameter (in the case of this test, two), the list must contain a list of possible values, e.g. [ [10, 20], [2, 3] ]. If None, a standard set is used. Optional, default: None.
    criterion : function
        Function to be applied to the set of p-values to obtain the best option. Optional, default: numpy.median.
    numProcesses : int
        Number of parallel tasks used in the evaluation. If -1, only one task is used. Optional, default: False.
    kwargs : 
        Other options to be passed to the function to obtain the p-values.

    Returns
    -------
    dictionary
        Set of best parameters.
    float
        Lowest obtained p-value.

    Raises
    ------
    ValueError
        If the parameters are not correct.
    """


    if paramSet is None:
        p1Set = [ 10, 20, 40 ]
        p2Set = [ 2, 3, 4, 5 ]
    else:
        p1Set = paramSet[ 0 ]
        p2Set = paramSet[ 1 ]

    bestPValue = 1.0
    bestParameters = []

    for pSet in product( p1Set, p2Set ):

        pV = np.zeros( ( len( tsSet ) ) )

        if numProcesses == -1:

            for k in range( len( tsSet ) ):
                pV[ k ] = GetPValue( tsSet[ k ], deltaT = pSet[ 0 ], granularity = pSet[ 1 ], **kwargs )[ 0 ]

        else:

            with Pool( processes = numProcesses ) as pool:

                async_result = []
                for k in range( len( tsSet ) ):
                    func = partial( GetPValue, TS = tsSet[ k ], deltaT = pSet[ 0 ], \
                                    granularity = pSet[ 1 ], **kwargs )
                    async_result.append( pool.apply_async( func ) )
                
                [result.wait() for result in async_result]
                for k in range( len( tsSet ) ):
                    pV[ k ] = async_result[ k ].get()[0]

        synthPV = criterion( pV )
        if synthPV < bestPValue:
            bestPValue = synthPV
            bestParameters = { 'deltaT': pSet[ 0 ], 'granularity': pSet[ 1 ] }

    return bestParameters, bestPValue





"""
BDS test, using the statsmodels implementation.

From:
    Broock, W. A., Scheinkman, J. A., Dechert, W. D., & LeBaron, B. (1996). 
    A test for independence based on the correlation dimension. Econometric reviews, 15(3), 197-235.
"""

import numpy as np
from statsmodels.tsa.stattools import bds



def GetPValue( TS, **kwargs ):


    """
    Get p-value according to the BDS test.

    Parameters
    ----------
    TS : numpy.array
        Time series to be analysed.
    max_dim : int
        The maximum embedding dimension. Optional, default: 2.
    distance : float
        Specifies the distance multiplier to use when computing the test. Optional, default: 1.5.

    Returns
    -------
    float
        p-value of the test.
    float
        statistic of the test.

    Raises
    ------
    ValueError
        If the parameters are not correct.
    """


    # Setting up the parameters

    max_dim = kwargs.get( 'max_dim', 2 )
    distance = kwargs.get( 'distance', 1.5 )


    # Checking the parameters

    if type( max_dim ) is not int:
        raise ValueError("max_dim is not an integer")
    if max_dim <= 0:
        raise ValueError("max_dim must be larger than 0")
    if type( distance ) is not float:
        raise ValueError("distance is not an float")
    if distance <= 0:
        raise ValueError("distance must be larger than 0")


    # Computing the test

    bds_stat, pV = bds( TS, max_dim = max_dim, epsilon = None, distance = distance )

    if pV.size > 1:
        bds_stat = bds_stat[ np.argmin( pV ) ]
        pV = np.min( pV )
    
    return float( pV ), float( bds_stat )



def GetStatistic( TS, **kwargs ):

    """
    Get the statistic according to the BDS test.

    Parameters
    ----------
    TS : numpy.array
        Time series to be analysed.
    max_dim : int
        The maximum embedding dimension. Optional, default: 2.
    distance : float
        Specifies the distance multiplier to use when computing the test. Optional, default: 1.5.

    Returns
    -------
    float
        statistic of the test.

    Raises
    ------
    ValueError
        If the parameters are not correct.
    """

    return GetPValue( TS, **kwargs )[ 1 ]





from .optionalNJIT import optional_njit

printWarnings = True



@optional_njit( cache=True, nogil=True )   
def warnUser( source: str, message: str ) -> None:
    global printWarnings
    if printWarnings:
        print( 'Warning from: ' + source )
        print( message )




import numpy as np
from ..optionalNJIT import optional_njit
from ..warningHandling import warnUser


@optional_njit( cache=True, nogil=True )            
def OrnsteinUhlenbeck( tsLen: int = 1000, x: float = 0.0, \
                       Omega: float = 0.1, avg: float = 0.0, \
                       delta: float = 1.0, \
                       randomInit: bool = False ) -> np.ndarray :
      
    """
    Generate a time series using the Ornstein-Uhlenbeck process.
    The Ornstein-Uhlenbeck process is defined as:
    .. math:: dx_t = \\theta (\\mu - x_t) dt + \\sigma dW_t
    where :math:`x_t` is the value at time step :math:`t`, :math:`\\theta` is the rate of mean reversion,
    :math:`\\mu` is the long-term mean, :math:`\\sigma` is the volatility, and :math:`dW_t` is a Wiener process.
    
    Parameters
    ----------
    tsLen : int
        Length of the time series to generate.
    x : float
        Initial value of the time series.
    Omega : float
        Rate of mean reversion.
    avg : float
        Long-term mean.
    delta : float
        Volatility.
    randomInit : bool
        If True, a random initial value is generated between 0.0 and 1.0.

    Returns
    -------
    np.ndarray
        Generated time series.
    """

    if randomInit:
        x = np.random.uniform( 0.0, 1.0 )

    if tsLen < 1:
        raise ValueError( "tsLen must be greater than 0" )
    if delta < 0.0:
        raise ValueError( "delta must be greater than 0" )

    TS = np.zeros( (tsLen) )
    TS[0] = x
                                          
    for i in range(1, tsLen):

        x_dot = np.random.normal( 0.0, delta )   
        x_dot = Omega * ( avg - TS[i - 1] ) + x_dot
        TS[i] += x_dot
    
    return TS





"""
Continuous Ordinal Patterns test

From:
    Zanin, M. (2023).
    Continuous ordinal patterns: Creating a bridge between ordinal analysis and deep learning.
    Chaos: An Interdisciplinary Journal of Nonlinear Science 33 (3).
"""

import numpy as np
from ..optionalNJIT import optional_njit
from scipy.stats import ks_2samp



@optional_njit( cache=True, nogil=True )
def _numba_normWindow( ts ):
	
	newTS = np.copy( ts )
	newTS -= np.min( newTS )
	newTS /= np.max( newTS )
	newTS = ( newTS - 0.5 ) * 2.0
	return newTS
	
		


@optional_njit( cache=True, nogil=True )            
def _numba_tranfTS( ts, patt ):
	
    tsL = ts.shape[0]
    W = np.zeros( ( tsL - patt.shape[0] + 1 ) )
	
    for t in range( tsL - patt.shape[0] + 1 ):
		
        for l in range( patt.shape[0] ):
            
            W[ t ] += np.abs( _numba_normWindow( ts[ t : ( t + patt.shape[0] ) ] )[ l ] - patt[ l ] )

        W[ t ] = W[ t ] / patt.shape[0] / 2.0
		
    return W



def _shader_tranfTS_Split( ts, patt ):

    from wgpu.utils.compute import compute_with_buffers
    m = np.size( ts )
    n = np.size( patt )

    A_shape = ( m )
    B_shape = ( n )

    bindings = {
        0: ts.astype(np.float32),
        1: patt.astype( np.float32 ),
        3: np.array(A_shape, dtype=np.uint32),
        4: np.array(B_shape, dtype=np.uint32),
    }

    shader_src = """
    @group(0) @binding(0)
    var<storage, read> A: array<f32>;
    @group(0) @binding(1)
    var<storage, read> B: array<f32>;
    @group(0) @binding(2)
    var<storage, read_write> C: array<f32>;

    @group(0) @binding(3)
    var<storage, read> A_shape: array<u32>;

    override n: u32;

    @compute @workgroup_size(1)
    fn main(@builtin(global_invocation_id) gid: vec3<u32>) {

        let m: u32 = A_shape[0];
        var distance: f32 = 0.;

        var j: u32 = gid.x;

        var minV: f32 = 999999.9;
        for (var i: u32 = 0; i < n; i++) {
            if A[ j + i ] < minV {
                minV = A[ j + i ];
            }
        }        

        var subW: array<f32, 100>;
        for (var i: u32 = 0; i < n; i++) {
            subW[ i ] = A[ j + i ] - minV;
        }

        var maxV: f32 = -999999.9;
        for (var i: u32 = 0; i < n; i++) {
            if subW[ i ] > maxV {
                maxV = subW[ i ];
            }
        }        

        for (var i: u32 = 0; i < n; i++) {
            subW[ i ] = subW[ i ] / maxV;
            subW[ i ] = ( subW[ i ] - 0.5 ) * 2.;
        }
        
        for (var i: u32 = 0; i < n; i++) {
            distance += abs( subW[ i ] - B[ i ] );
        }
        
        C[ j ] = distance;

        return;
    }
    """

    out = compute_with_buffers(
        input_arrays=bindings,
        output_arrays={2: (m, "f")},
        shader=shader_src,
        constants={'n': n},
        n=(m, 1, 1), 
    )

    C = np.frombuffer(out[2], dtype=np.float32) / n / 2.
    return C



def _shader_tranfTS( TS, patt ):
    C = np.zeros( (0) )
    if np.size( TS ) < 60000:
        C = _shader_tranfTS_Split( TS, patt )[ : - patt.shape[0] + 1 ]
    else:
        startOffset = 0
        while startOffset + 60000 + patt.shape[0] - 1 < np.size( TS ):
            ax = _shader_tranfTS_Split( \
                TS[ startOffset : startOffset + 60000 + patt.shape[0] - 1 ], \
                patt )[ : - patt.shape[0] + 1 ]
            C = np.hstack( ( C, ax ) )
            startOffset += 60000
        ax = _shader_tranfTS_Split( TS[ startOffset : ], patt )[ : - patt.shape[0] + 1 ]
        C = np.hstack( ( C, ax ) )

    return C





def _evaluateDistance( patt, TS1, TS2, useShaders ):
    
    patt = _numba_normWindow( patt )
    
    d_0 = np.empty( ( 0, 1 ) )
    d_r = np.empty( ( 0, 1 ) )

    TS = TS1
    if useShaders:
        w = _shader_tranfTS( TS, patt )
    else:
        w = _numba_tranfTS( TS, patt )
    w[ np.isnan( w ) ] = 0.0
    d_0 = w
    
    TS = TS2
    if useShaders:
        w = _shader_tranfTS( TS, patt )
    else:
        w = _numba_tranfTS( TS, patt )
    w[ np.isnan( w ) ] = 0.0
    d_r = w

    allRes = ks_2samp( d_0, d_r )
       
    return allRes[ 1 ], allRes[ 0 ]



def GetPValue( TS, **kwargs ):

    """
    Get p-value according to the Continuous Ordinal Pattern test.

    Parameters
    ----------
    TS : numpy.array
        Time series to be analysed.
    pSize : int
        Size of the pattern. Optional, default: 3.
    numIters : int
        Number of random patterns to be tested. Optional, default: function of the pattern size.
    useShaders : bool
        Whether to use shaders or only CPU computation. Optional, default: False.

    Returns
    -------
    float
        p-value of the test.
    float
        statistic of the test.

    Raises
    ------
    ValueError
        If the parameters are not correct.
    """


    # Setting up the parameters

    pSize = kwargs.get( 'pSize', 3 )
    numIters = kwargs.get( 'numIters', int( 10 ** np.sqrt( pSize - 2 ) ) )
    useShaders = kwargs.get( 'useShaders', False )
     

    # Checking the parameters

    if type( pSize ) is not int:
        raise ValueError("pSize is not an integer")
    if pSize <= 2:
        raise ValueError("pSize must be larger than 2")
    if type( numIters ) is not int:
        raise ValueError("numIters is not an integer")
    if numIters < 1:
        raise ValueError("numIters must be larger than zero")
    if type( useShaders ) is not bool:
        raise ValueError("useShaders is not a bool")
    if useShaders and pSize > 100:
        raise ValueError("pSize cannot be greater than 100 when using shaders")


    # Compute the test

    bestD = np.zeros( ( numIters ) )
    bestS = np.zeros( ( numIters ) )
    rTS = TS[::-1]

    for k in range( numIters ):
        x0 = np.random.uniform( 0.0, 1.0, (pSize) )
        bestD[ k ], bestS[ k ] = _evaluateDistance( x0, np.copy( TS ), np.copy( rTS ), useShaders )

    return np.min( bestD ), bestS[ np.argmin( bestD ) ]




def GetStatistic( TS, **kwargs ):

    """
    Get the statistic according to the Continuous Ordinal Pattern test.

    Parameters
    ----------
    TS : numpy.array
        Time series to be analysed.
    pSize : int
        Size of the pattern. Optional, default: 3.
    numIters : int
        Number of random patterns to be tested. Optional, default: function of the pattern size.
    useShaders : bool
        Whether to use shaders or only CPU computation. Optional, default: False.

    Returns
    -------
    float
        statistic of the test.

    Raises
    ------
    ValueError
        If the parameters are not correct.
    """

    return GetPValue( TS, **kwargs )[ 1 ]





"""
Permutation patterns test

From:
    Zanin, M., Rodríguez-González, A., Menasalvas Ruiz, E., & Papo, D. (2018).
    Assessing time series reversibility through permutation patterns. Entropy, 20(9), 665.
"""


try:
    from scipy.stats import binom_test
except:
    from scipy import stats
    binomtest = stats.binomtest
    pass

from itertools import permutations
import numpy as np
from ..optionalNJIT import optional_njit



@optional_njit( cache=True, nogil=True )
def _numba_GetSinglePermutation( TS ):
    
    order = np.argsort( TS )

    return order
    
    
    
@optional_njit( cache=True, nogil=True )
def _numba_GetPermProb( TS, wLength, allPatterns ):
    
    pProb = np.zeros( ( allPatterns.shape[ 0 ] ) )    
    numPatt = allPatterns.shape[ 0 ]
    
    tsLength = TS.shape[0]
    
    for k in range(0, tsLength - wLength ):
        
        subTS = TS[ k : (k + wLength) ]
        pattern = np.argsort( subTS )

        offset = -1        
        for l in range( numPatt ):
            if np.all( allPatterns[l, :] == pattern ):
                offset = l
                break
            
        if offset == -1:
            allPatterns[ numPatt, : ] = pattern
            pProb[ numPatt ] = 1.0
            numPatt += 1
        else:
            pProb[ offset ] += 1.0
            
    allPatterns = allPatterns[ :numPatt, : ]
    pProb = pProb[ :numPatt ]
         
    totPatterns = np.sum( pProb )
    for k in range( numPatt ):
        pProb[k] /= totPatterns
    
    return ( allPatterns, pProb )
    


def _BinomialTestCompatibility( x, p = 0.5 ):

    try:
        if np.sum( x ) == 0: return 0.
        
        pV = binomtest( x[ 0 ], n = np.sum( x ), p = p ).pvalue
    except:
        pV = binom_test( x, p = p )
        pass

    return pV
    

    

def GetPValue( TS, **kwargs ):

    """
    Get p-value according to the Permutation Pattern test.

    Parameters
    ----------
    TS : numpy.array
        Time series to be analysed.
    pSize : int
        Size of the permutation pattern, i.e. embedding dimension. Optional, default: 3.

    Returns
    -------
    float
        p-value of the test.
    float
        statistic of the test.

    Raises
    ------
    ValueError
        If the parameters are not correct.
    """
        

    # Setting up the parameters

    pSize = kwargs.get( 'pSize', 3 )


    # Checking the parameters

    if type( pSize ) is not int:
        raise ValueError("pSize is not an integer")
    if pSize <= 2:
        raise ValueError("pSize must be larger than two")


    # Compute the test

    res0, res1 = _numba_GetPermProb( np.ravel(TS), pSize, np.array( list( permutations( range( pSize ) ) ) ) )
    
    asym = []
    
    for k in range( len( res0 ) ):
        for l in range( k + 1, len( res0 ) ):
        
            if np.all( res0[ k ] == res0[ l ][::-1] ):
                
                numPatterns = np.size( TS, 0 ) - 3
                x = [ int( res1[ k ] * numPatterns ),  int( res1[ l ] * numPatterns ) ]
                pV = _BinomialTestCompatibility( x )
                asym.append( pV ) 
                break

    asym = np.array( asym )
    
    if np.size( asym ) == 0:
        return 1.0, 0.0
        
    asym = np.min( asym ) * pSize
    if asym > 1.0: asym = 1.0

    return asym, 0.0



    

def GetStatistic( TS, **kwargs ):

    """
    Get the statistics according to the Permutation Pattern test.
    Note: this function is yet to be implemented!

    Parameters
    ----------
    TS : numpy.array
        Time series to be analysed.
    pSize : int
        Size of the permutation pattern, i.e. embedding dimension. Optional, default: 3.

    Returns
    -------
    float
        statistic of the test.

    Raises
    ------
    ValueError
        If the parameters are not correct.
    """
        
    return GetPValue( TS, **kwargs )[ 1 ]




"""
Functions to generate random walks on networks, which can be interpreted as symbolic sequences.
"""


import numpy as np
from ..optionalNJIT import optional_njit
from ..warningHandling import warnUser



def UnbiasedRandomWalk( numNodes: int, lPath: int ) -> np.ndarray :

    """
    Generate a symbolic time series using an unbiased random walk on a network.
    
    Parameters
    ----------
    numNodes : int
        Number of nodes in the network, hence of symbols.
    lPath : float
        Length of the time series (path) to be generated.

    Returns
    -------
    np.ndarray
        Generated time series.
    """

    if numNodes < 1:
        raise ValueError( "numNodes must be greater than 0" )
    if lPath < 1:
        raise ValueError( "lPath must be greater than 0" )

    Path = [ ]

    for i in range( lPath ):
        Path.append( np.random.randint( 0, numNodes ) )

    return np.array( Path, dtype = int )



def BiasedRandomWalk( numNodes, lPath, alpha ) -> np.ndarray :

    """
    Generate a symbolic time series using an biased random walk on a network.
    The biases are generated by the links' weights. Additionally, the parameter
    :math:`\\alpha` allows to introduce a random reset.
    
    Parameters
    ----------
    numNodes : int
        Number of nodes in the network, hence of symbols.
    lPath : float
        Length of the time series (path) to be generated.
    alpha : float
        Probability of jumping to a random node. For :math:`\\alpha = 1` an unbiased walk is recovered.

    Returns
    -------
    np.ndarray
        Generated time series.
    """
    if numNodes < 1:
        raise ValueError( "numNodes must be greater than 0" )
    if lPath < 1:
        raise ValueError( "lPath must be greater than 0" )
    if alpha < 0.0 or alpha > 1.0:
        warnUser( "TimeSeriesGeneration.SymbolicRandomWalk.BiasedRandomWalk", \
                  "alpha should be between 0.0 and 1.0" )

    W = np.random.uniform( 0, 1, size = ( numNodes, numNodes ) )
    for n1 in range( numNodes ):
        W[ n1, n1 ] = 0.
        for n2 in range( n1 + 1, numNodes ):
            W[ n1, n2 ] = 1. - W[ n2, n1 ]

    for n1 in range( numNodes ):
        W[ n1, : ] = W[ n1, : ] / np.sum( W[ n1, : ] )

    Path = [ ]

    Path.append( np.random.randint( 0, numNodes ) )
    for i in range( 1, lPath ):
        n1 = Path[ -1 ]
        p = ( 1. - alpha ) * W[ n1, : ] + alpha
        p /= np.sum( p )
        n2 = np.random.choice( numNodes, p = p )
        Path.append( n2 )

    return np.array( Path, dtype = int )




import numpy as np
from ..optionalNJIT import optional_njit
from ..warningHandling import warnUser


@optional_njit( cache=True, nogil=True )   
def ST( offset, period, w ):

    newOff = np.mod( offset, period ) / period

    refValuesX = [ 0.0, w, 1.0 ]
    refValuesY = [ 0.0, 1.0, 0.0 ]

    value = np.interp( newOff, refValuesX, refValuesY )
    return value



@optional_njit( cache=True, nogil=True )   
def aWeierstrass( tsLen: int = 1000, period: int = 100, \
                  numHarms: int = 50, w: float = 0.2 ) -> np.ndarray :
        
    """
    Generate a time series using the Weierstrass function. See:
    Burykin, A., Costa, M. D., Peng, C. K., Goldberger, A. L., & Buchman, T. G. (2011). Generating signals with multiscale time irreversibility: the asymmetric weierstrass function. Complexity, 16(4), 29-38.
    
    Parameters
    ----------
    tsLen : int
        Length of the time series to generate.
    period : int
        Period of the time series.
    numHarms : int
        Number of harmonics to include in the time series.
    w : float
        Asymmetry factor of all harmonics.

    Returns
    -------
    np.ndarray
        Generated time series.
    """

    if tsLen < 1:
        raise ValueError( "tsLen must be greater than 0" )
    if period < 1:
        raise ValueError( "period must be greater than 0" )
    if numHarms < 1:
        raise ValueError( "numHarms must be greater than 0" )
    if w < 0.0 or w > 1.0:
        raise ValueError( "w must be between 0.0 and 1.0" )

    TS = np.zeros( ( tsLen ) )

    for k in range( 1, numHarms + 1 ):

        tPeriod = period / k
        tAmpl = 1 / ( k / 2.0 )
        TS += ST( np.arange( tsLen ), tPeriod, w ) * tAmpl

    TS /= np.max( TS )

    return TS






"""
Utility functions to downsampling a time series.
"""

import numpy as np
from scipy.signal import decimate
from ..warningHandling import warnUser



def __ParametersCheck( tau ):

    if tau < 1:
        raise ValueError( "Tau must be greater than 0" )
    


def OneEveryN( TS: np.ndarray, tau: int ) -> np.ndarray :

    """
    Downsample a time series by keeping one every tau observations.
    
    Parameters
    ----------
    TS : np.ndarray
        Input time series.
    tau : int
        Downsampling factor.

    Returns
    -------
    np.ndarray
        Generated time series.
    """

    __ParametersCheck( tau )
    
    return np.copy( TS[ :: tau ] )



def Downsampling_Avg( TS: np.ndarray, tau: int ) -> np.ndarray :

    """
    Downsample a time series by extracting sub-windows of size tau, and substituting their values with the corresponding average.
    
    Parameters
    ----------
    TS : np.ndarray
        Input time series.
    tau : int
        Downsampling factor.

    Returns
    -------
    np.ndarray
        Generated time series.
    """

    __ParametersCheck( tau )
    
    newL = int( np.size( TS ) / tau )
    newTS = np.zeros( ( newL ) )
    for k in range( newL ):
        newTS[ k ] = np.mean( TS[ ( k * tau ) : ( (k+1) * tau ) ] )
    return newTS



def Decimate( TS: np.ndarray, tau: int ) -> np.ndarray :

    """
    Downsample a time series through a decimation process; see https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.decimate.html.
    
    Parameters
    ----------
    TS : np.ndarray
        Input time series.
    tau : int
        Downsampling factor.

    Returns
    -------
    np.ndarray
        Generated time series.
    """

    __ParametersCheck( tau )
        
    return decimate( np.copy( TS ), tau )





"""
Visibility Graph test

From:
    Lacasa, L., Nunez, A., Roldán, É., Parrondo, J. M., & Luque, B. (2012).
    Time series irreversibility: a visibility graph approach. The European Physical Journal B, 85(6), 1-11.
"""

import numpy as np
from scipy.stats import epps_singleton_2samp, ks_2samp
from scipy.stats import entropy
from ..optionalNJIT import optional_njit




@optional_njit( cache=True, nogil=True )            
def _numba_GetOneDegree( TS, offset, horizon ):
    
    tsLength = TS.shape[0]
    
    tempDegree = 1.0

    if horizon == -1:
        maxDistance = tsLength
    else:
        maxDistance = offset + horizon
        if maxDistance > tsLength:
            maxDistance = tsLength
    
    if TS[ offset + 1 ] >= TS[ offset ]: return tempDegree

    lastMax = TS[ offset + 1 ]

    for k in range( offset + 2, maxDistance ):
        
        if TS[ k ] >= lastMax:
            lastMax = TS[ k ]
            tempDegree += 1

        if lastMax > TS[ offset ]:
            break
            
    return tempDegree
    


def _GetVGDivergence( dD, rD ):
        
    while True:
        if dD[ -1 ] == 0 and rD[ -1 ] == 0:
            dD = dD[ :-1 ]
            rD = rD[ :-1 ]
            continue
        break
    
    dD /= np.sum( dD )
    rD /= np.sum( rD )
    en = entropy( dD + 0.000001, qk = rD + 0.000001 )
    
    return en




@optional_njit( cache=True, nogil=True )            
def _numba_GetAllDegrees( TS, horizon ):

    tsLength = TS.shape[0]
    allDegrees = np.zeros( (tsLength - 2) )
    
    for k in range( tsLength - 2 ):        
        allDegrees[k] = _numba_GetOneDegree( TS, k, horizon )


    degreeDistr = np.zeros( (tsLength) )
    
    for k in range( tsLength ):    
        degreeDistr[ k ] = np.sum( allDegrees == k )

        
    return allDegrees, degreeDistr




@optional_njit( cache=True, nogil=True )            
def _numba_GetAllDegrees_Parallel( TS, horizon ):

    tsLength = TS.shape[0]
    allDegrees = np.zeros( (tsLength - 2) )
    
    for k in range( tsLength - 2 ):        
        allDegrees[k] = _numba_GetOneDegree( TS, k, horizon )


    degreeDistr = np.zeros( (tsLength) )
    
    for k in range( tsLength ):    
        degreeDistr[ k ] = np.sum( allDegrees == k )

        
    return allDegrees, degreeDistr




    
def GetPValue( TS: np.array, **kwargs ):

    """
    Get p-value according to the Visibility Graph test.

    Parameters
    ----------
    TS : numpy.array
        Time series to be analysed.
    horizon : int
        Maximum distance for detecting links. If '-1', the whole time series is analysed. Optional, default: -1.
    parallel : bool
        Activate parallelisation - not yet implemented. Optional, default: False.

    Returns
    -------
    float
        p-value of the test.
    float
        statistic of the test.

    Raises
    ------
    ValueError
        If the parameters are not correct.
    """


    # Setting up the parameters

    horizon = kwargs.get( 'horizon', -1 )
    parallel = kwargs.get( 'parallel', False )


    # Checking the parameters

    if type( horizon ) is not int:
        raise ValueError("horizon is not an integer")
    if horizon < -1:
        raise ValueError("horizon must be larger than 0, or -1")


    # Computing the degree distributions

    if parallel:
        allD, dD = _numba_GetAllDegrees_Parallel( TS, horizon )
        allR, rD = _numba_GetAllDegrees_Parallel( TS[ ::-1 ], horizon )
    else:
        allD, dD = _numba_GetAllDegrees( TS, horizon )
        allR, rD = _numba_GetAllDegrees( TS[ ::-1 ], horizon )


    # Extract the p-value

    pV = 1.0
    try:
        pV = epps_singleton_2samp( allD,  allR )[1]
    except:
        pass

    if pV == 1.0:
        try:
            pV = ks_2samp( allD,  allR )[1]
        except:
            pass
    
    div = _GetVGDivergence( dD, rD )
        
    return pV, div



def GetStatistic( TS, **kwargs ):

    """
    Get the statistic of the Visibility Graph test.

    Parameters
    ----------
    TS : numpy.array
        Time series to be analysed.
    horizon : int
        Maximum distance for detecting links. If '-1', the whole time series is analysed. Optional, default: -1.
    parallel : bool
        Activate parallelisation - not yet implemented. Optional, default: False.

    Returns
    -------
    float
        statistic of the test.

    Raises
    ------
    ValueError
        If the parameters are not correct.
    """

    return GetPValue( TS, **kwargs )[ 1 ]





"""
Micro-scale trends test

From:
    Zanin, M. (2021). Assessing time series irreversibility through micro-scale trends.
    arXiv preprint arXiv:2108.06272.
"""

import numpy as np
from scipy import stats
from scipy.stats import skew
from ..optionalNJIT import optional_njit




@optional_njit( cache=True, nogil=True )
def _numba_OneSlope( Y, wSize ):

    X = np.arange( wSize )
    xMean = np.mean( X )
    yMean = np.mean( Y )
    X = X - xMean
    Y = Y - yMean

    res = 0.0
    for k in range( wSize ):
        res += X[k] * Y[k]

    res2 = 0.0
    for k in range( wSize ):
        res2 += X[k] * X[k]

    return res / res2



@optional_njit( cache=True, nogil=True )
def _numba_SlopeIrreversibility_aux( TS, wSize ):

    numSamples = TS.shape[0]
    allSlopes = np.zeros((numSamples - wSize))

    for k in range(numSamples - wSize):
        subTS = TS[k: k + wSize]
        allSlopes[k] = _numba_OneSlope( subTS, wSize )

    return allSlopes





@optional_njit( cache=True, nogil=True )
def _numba_SlopeIrreversibility_aux_2( TS, wSize2 ):

    tsLen = TS.shape[0]
    TSStd = np.zeros ( (tsLen - 2 * wSize2) )
    for k in range ( wSize2, tsLen - wSize2 ):
        TSStd[ k - wSize2 ] = np.std ( TS[ k - wSize2 : k + wSize2 ] )

    return TSStd





@optional_njit( cache=True, nogil=True )
def _numba_SlopeIrreversibility_aux_3( TS, wSize2 ):

    tsLen = TS.shape[0]
    TSStd = np.zeros ( (tsLen - 2 * wSize2) )
    for k in range ( wSize2, tsLen - wSize2 ):

        temp = ( TS[ k - wSize2 : k + wSize2 ] - np.mean( TS[ k - wSize2 : k + wSize2 ] ) )
        temp = np.power( temp, 3.0 )
        temp2 = np.sum( temp )
        temp2 = temp2 / float( tsLen - 2 * wSize2 )

        tempv = np.std( TS[ k - wSize2 : k + wSize2 ] )
        tempv = np.power( tempv, 3.0 )
        temp3 = temp2 / tempv
        TSStd[ k - wSize2 ] = temp3

    return TSStd




@optional_njit( cache=True, nogil=True )
def _numba_SlopeIrreversibility_aux_3b( TS, wSize2 ):

    tsLen = TS.shape[0]
    TSStd = np.zeros ( (tsLen - 2 * wSize2) )
    for k in range ( wSize2, tsLen - wSize2 ):
        TSStd[ k - wSize2 ] = skew( TS[ k - wSize2 : k + wSize2 ] )

    return TSStd





def GetPValue( TS, **kwargs ):

    """
    Get p-value according to the MSTrends test.

    Parameters
    ----------
    TS : numpy.array
        Time series to be analysed.
    wSize : int
        Size of the small window. Optional, default: 2.
    wSize2 : int
        Size of the large window. Optional, default: 20.

    Returns
    -------
    float
        p-value of the test.
    float
        statistic of the test.

    Raises
    ------
    ValueError
        If the parameters are not correct.
    """


    # Setting up the parameters

    wSize = kwargs.get( 'wSize', 2 )
    wSize2 = kwargs.get( 'wSize2', 20 )


    # Checking the parameters

    if type( wSize ) is not int:
        raise ValueError("wSize is not an integer")
    if wSize <= 0:
        raise ValueError("wSize must be larger than 0")
    if type( wSize2 ) is not int:
        raise ValueError("wSize2 is not an integer")
    if wSize2 <= 0:
        raise ValueError("wSize2 must be larger than 0")


    # Calculate the test

    allSlopes = _numba_SlopeIrreversibility_aux( TS, wSize )
    res1 = stats.ks_2samp(allSlopes, -allSlopes)

    res2 = [ 0.0, 1.0 ]
    try:
        TSStd = _numba_SlopeIrreversibility_aux_2( TS, wSize2 )
        allSlopes2 = _numba_SlopeIrreversibility_aux( TSStd, wSize )
        res2 = stats.ks_2samp(allSlopes2, -allSlopes2)
    except:
        pass

    res3 = [ 0.0, 1.0 ]
    try:
        TSSkew = _numba_SlopeIrreversibility_aux_3( TS, wSize2 )
        allSlopes3 = _numba_SlopeIrreversibility_aux( TSSkew, wSize )
        res3 = stats.ks_2samp(allSlopes3, -allSlopes3)
    except:
        pass

    p_value = np.min( [ res1[1], res2[1], res3[1] ] ) * 3.0
    stat = np.max( [ res1[0], res2[0], res3[0] ] )

    if p_value > 1.0: p_value = 1.0

    return p_value, stat






def GetStatistic( TS, **kwargs ):

    """
    Get the statistic according to the MSTrends test.

    Parameters
    ----------
    TS : numpy.array
        Time series to be analysed.
    wSize : int
        Size of the small window. Optional, default: 2.
    wSize2 : int
        Size of the large window. Optional, default: 20.

    Returns
    -------
    float
        statistic of the test.

    Raises
    ------
    ValueError
        If the parameters are not correct.
    """

    return GetPValue( TS, **kwargs )[ 1 ]

    



"""
Function to generate a time series using the linear congruential generator.
"""

import numpy as np
from ..optionalNJIT import optional_njit
from ..warningHandling import warnUser


@optional_njit( cache=True, nogil=True )   
def congruential( tsLen: int = 1000, x: float = 0.0, A: int = 7141, \
                  B: int = 54773, C: int = 259200, randomInit: bool = False ) -> np.ndarray :

    """
    Generate a time series using a linear congruential generator.
    
    Parameters
    ----------
    tsLen : int
        Length of the time series to generate.
    x : float
        Initial value of the time series.
    A : float
        Parameter A of the LCG.
    B : float
        Parameter B of the LCG.
    C : float
        Parameter C of the LCG.
    randomInit : bool
        If True, a random initial value is generated for x.

    Returns
    -------
    np.ndarray
        Generated time series.
    """

    if randomInit:
        x = np.random.uniform( 0, C )

    if tsLen < 1:
        raise ValueError( "tsLen must be greater than 0" )

    TS = np.zeros( ( tsLen ) )
    for k in range( 1, tsLen ):
        TS[ k ] = np.mod( A * TS[ k - 1 ] + B, C )
        
    TS /= C
    
    return TS




import pytest
from pytest import approx

import numpy as np
import copy

import irreversibility as irr



@pytest.mark.parametrize("length", [ 100, 1000, 10000 ])
def test_TS_Does_Not_Change( length ):

    TS = np.random.uniform( 0., 1., (length) )
    TS2 = copy.deepcopy( TS )

    irr.Metrics.Diks.GetPValue( TS )
    assert np.sum( TS - TS2 ) == 0, (
        "Time series has been modified"
    )



@pytest.mark.parametrize("length", [ 100, 1000, 10000 ])
def test_GetPValue( length ):
    """Test the functions for calculating the p-value"""

    TS = np.random.uniform( 0., 1., (length) )

    with pytest.raises(ValueError):
        irr.Metrics.Diks.GetPValue( TS, embD = 1. )
    with pytest.raises(ValueError):
        irr.Metrics.Diks.GetPValue( TS, embD = -1 )

    with pytest.raises(ValueError):
        irr.Metrics.Diks.GetPValue( TS, dVar = -1. )

    irr.Metrics.Diks.GetPValue( TS )



@pytest.mark.parametrize("length", [ 100, 1000, 10000 ])
def test_GetStatistic( length ):
    """Test the functions for calculating the p-value"""

    TS = np.random.uniform( 0., 1., (length) )

    result = irr.Metrics.Diks.GetStatistic( TS )
    assert isinstance( result, float )




def test_Optimisation():
    """Test the functions for optimising the parameters"""

    from irreversibility.ParameterOptimisation.Opt_Diks import Optimisation

    tsSet = []
    for k in range( 10 ):
        tsSet.append( np.random.uniform( 0., 1., (645) ) )

    Optimisation( tsSet, paramSet = None, criterion = np.median, numProcesses = -1 )





"""
Costa index test

From:
    Costa, M.; Goldberger, A.L.; Peng, C.K.
    Broken asymmetry of the human heartbeat: loss of time irreversibility in aging and disease.
    Physical review letters 2005, 95, 198102.
"""

from scipy.stats import norm
import numpy as np




def _Get_A_For_Rnd_Series( TS, tau, numRndReps = 100 ):

    a = []
    
    for k in range( numRndReps ):
        tTS = np.random.permutation( np.copy( TS ) )
        a.append( _Get_A( tTS, tau = tau ) )
    
    return a



def _Get_A( TS, tau ):

    dTS = TS[ tau: ] - TS[ :-tau ]
    a1 = np.sum( dTS > 0 )
    a2 = np.sum( dTS < 0 )
    return np.abs( a1 - a2 ) / ( a1 + a2 )



def GetPValue( TS, numRndReps = 100, **kwargs ):

    """
    Get p-value according to the Costa test.

    Parameters
    ----------
    TS : numpy.array
        Time series to be analysed.
    tau : int
        Lag used in the computation of the function. Optional, default: 1.
    numRndReps : int
        Number of random shuffled time series used to estimate the p-value. Optional, default: 100.
    pValueMethod : string
        Method used to estimate the p-value, including 'proportion' and 'z-score'. Optional, default: 'z-score'.

    Returns
    -------
    float
        p-value of the test.
    float
        statistic of the test.

    Raises
    ------
    ValueError
        If the parameters are not correct.
    """


    # Setting up the parameters

    tau = kwargs.get( 'tau', 1 )
    numRndReps = kwargs.get( 'numRndReps', 100 )
    pValueMethod = kwargs.get( 'pValueMethod', 'z-score' )


    # Checking the parameters

    if type( tau ) is not int:
        raise ValueError("tau is not an integer")
    if tau <= 0:
        raise ValueError("tau must be larger than 0")
    if type( numRndReps ) is not int:
        raise ValueError("numRndReps is not an integer")
    if numRndReps < 0:
        raise ValueError("numRndReps must be zero or positive")
    if pValueMethod not in [ 'proportion', 'z-score' ]:
        raise ValueError("pValueMethod not recognised")


    # Computing the psi for the original time series

    a = _Get_A( TS, tau )
    
    if numRndReps <= 0:
        return 1.0, a


    # Extract the p-value

    pValue = 1.0
    rndA = _Get_A_For_Rnd_Series( TS, tau = tau, numRndReps = numRndReps )

    if pValueMethod == 'z-score':
        z_score = ( a - np.mean( rndA ) ) / np.std( rndA )
        pValue = norm.sf( z_score )
    if pValueMethod == 'proportion':
        pValue = np.sum( np.abs( a ) < np.abs( np.array( rndA ) ) ) / numRndReps

    return pValue, a




def GetStatistic( TS, **kwargs ):

    """
    Get the statistic according to the Costa test.

    Parameters
    ----------
    TS : numpy.array
        Time series to be analysed.
    tau : int
        Lag used in the computation of the function. Optional, default: 1.

    Returns
    -------
    float
        statistic of the test.

    Raises
    ------
    ValueError
        If the parameters are not correct.
    """
    
    return GetPValue( TS, numRndReps = 0, **kwargs )[ 1 ]







"""
Gaspard's estimation of the entropy production for symbolic time series

From:
    Gaspard, P. (2004).
    Time-reversed dynamical entropy and irreversibility in Markovian random processes.
    Journal of statistical physics, 117(3), 599-615.
"""


import numpy as np
import itertools
from scipy.stats import norm
from ..optionalNJIT import optional_njit
from ..warningHandling import warnUser




def _shader_OneSequence( currentS, sequence, n ):

    from wgpu.utils.compute import compute_with_buffers
    m = np.size( sequence )

    A_shape = ( m )
    B_shape = ( n )

    newTS = sequence.astype(np.uint32)

    bindings = {
        0: newTS,
        1: currentS.astype( np.uint32 ),
        3: np.array(A_shape, dtype=np.uint32),
        4: np.array(B_shape, dtype=np.uint32),
    }

    shader_src = """
    @group(0) @binding(0)
    var<storage, read> A: array<u32>;
    @group(0) @binding(1)
    var<storage, read> B: array<u32>;
    @group(0) @binding(2)
    var<storage, read_write> C: array<u32>;

    @group(0) @binding(3)
    var<storage, read> A_shape: array<u32>;

    override n: u32;

    @compute @workgroup_size(1)
    fn main(@builtin(global_invocation_id) gid: vec3<u32>) {

        let m: u32 = A_shape[0];
        var sum1: u32 = 0;
        var sum2: u32 = 0;

        var j: u32 = gid.x;

        C[ j ] = 0;
        C[ j + m ] = 0;

        var counter: u32 = 0;
        for (var i: u32 = 0; i < n; i++) {
            if A[ j + i ] == B[ i ] {
                counter += 1;
            } 
        }
        if counter == n {
            C[ j ] = 1;
        }

        counter = 0;
        for (var i: u32 = 0; i < n; i++) {
            if A[ j + i ] == B[ n - i - 1 ] {
                counter += 1;
            } 
        }
        if counter == n {
            C[ j + m ] = 1;
        }

        return;
    }
    """


    out = compute_with_buffers(
        input_arrays=bindings,
        output_arrays={2: (2*m, "I")},
        shader=shader_src,
        constants={'n': n},
        n=(m-n, 1, 1), 
    )

    C = np.frombuffer(out[2], dtype=np.uint32)
    return ( np.sum( C[ :m ] ), np.sum( C[ m: ] ) )



def _shader_OneSequence_Split( currentS, sequence, n ):

    C = np.array( [ 0., 0. ] )
    if np.size( sequence ) < 60000:
        C += _shader_OneSequence( currentS, sequence, n )
    else:
        startOffset = 0
        while startOffset + 60000 + n < np.size( sequence ):
            C += _shader_OneSequence( currentS, sequence[ startOffset : startOffset + 60000 + n ], n )
            startOffset += 60000
        C += _shader_OneSequence( currentS, sequence[ startOffset : ], n )

    return C




@optional_njit( cache=True, nogil=True )  
def _numba_OneSequence( currentS, sequence, n ):

    count = 0
    for k in range( sequence.shape[0] - n ):
        if np.all( currentS == sequence[ k : (k+n) ] ):
            count += 1

    count2 = 0
    for k in range( sequence.shape[0] - n ):
        if np.all( currentS[::-1] == sequence[ k : (k+n) ] ):
            count2 += 1
            
    return count, count2



def EntropyProduction( sequence, n: int = 3, useShaders: bool = False ):

    """
    Calculate the entropy production, by comparing the frequency of sequences of symbols, and the frequency of their time-reversed versions.
    
    Parameters
    ----------
    sequence : np.ndarray
        Sequence of symbols (or time series) to be analysed. Must be composed of discrete values, i.e. integers.
    n : int
        Length of the sequences of symbols to be analysed. Optional, default: 1.
    useShaders : bool
        Whether to use shaders or only CPU computation. Optional, default: False. 

    Returns
    -------
    float
        Produced entropy.
    """

    symbols = np.unique( sequence )

    H = 0.
    H_r = 0.

    for currentS in itertools.permutations( symbols, r = n ):

        ax = np.array( currentS )

        if useShaders:
            count, count2 = _shader_OneSequence_Split( ax, sequence, n )
        else:
            count, count2 = _numba_OneSequence( ax, sequence, n )

        count /= sequence.shape[0] - n
        count2 /= sequence.shape[0] - n
        
        if count > 0:
            H -= count * np.log( count )

        if count2 > 0:
            H_r -= count * np.log( count2 )

    EntropyProd = ( H_r - H ) / float( n )

    return EntropyProd



def GetPValue( TS, **kwargs ):

    """
    Calculate the Gaspard's test. The p-value is obtained by comparing the obtained entropy production against a set of randomly shuffled time series.
    
    Parameters
    ----------
    TS : np.ndarray
        Sequence of symbols (or time series) to be analysed. Must be composed of discrete values, i.e. integers.
    n : int
        Length of the sequences of symbols to be analysed. Optional, default: 3.
    numRndReps : int
        Number of random shuffled time series used to estimate the p-value. Optional, default: 100.
    useShaders : bool
        Whether to use shaders or only CPU computation. Optional, default: False.
    pValueMethod : string
        Method used to estimate the p-value, including 'proportion' and 'z-score'. Optional, default: 'proportional'.

    Returns
    -------
    float
        Obtained p-value.
    float
        Produced entropy.
    """

    # Setting up the parameters

    n = kwargs.get( 'n', 3 )
    numRndReps = kwargs.get( 'numRndReps', 100 )
    useShaders = kwargs.get( 'useShaders', False )
    pValueMethod = kwargs.get( 'pValueMethod', 'proportion' )


    # Checking the parameters

    if type( TS ) is not np.ndarray:
        raise ValueError( "TS must be a numpy array" )
    if len( np.unique( TS ) ) == np.size( TS ):
        warnUser( "DiscreteMetrics.Gaspard.GetPValue", \
                  "TS doesn't seem to be composed of discrete values" )
    if type( n ) is not int:
        raise ValueError("n is not an integer")
    if n <= 0:
        raise ValueError("n must be larger than 0")
    if n >= np.size( TS ):
        raise ValueError( "n must be smaller than the length of TS" )
    if type( numRndReps ) is not int:
        raise ValueError("numRndReps is not an integer")
    if numRndReps < 0:
        raise ValueError("numRndReps must be zero or positive")
    if type( useShaders ) is not bool:
        raise ValueError("useShaders is not a bool")
    if pValueMethod not in [ 'proportion', 'z-score' ]:
        raise ValueError("pValueMethod not recognised")
        

    # Computing the entropy production

    DS = EntropyProduction( TS, n = n, useShaders = useShaders )

    if numRndReps <= 0:
        return 1.0, DS

    
    # Compute the entropy production for the shuffled versions of the time series

    dsRnd = []
    for k in range( numRndReps ):
        dsRnd.append( EntropyProduction( np.random.permutation( TS ), \
                                         n = n, useShaders = useShaders ) )


    # Extract the p-value

    pValue = 1.0

    if pValueMethod == 'z-score':
        z_score = ( DS - np.mean( dsRnd ) ) / np.std( dsRnd )
        pValue = norm.sf( z_score )
    if pValueMethod == 'proportion':
        pValue = np.sum( DS < np.abs( np.array( dsRnd ) ) ) / numRndReps

    return ( pValue, DS )



def GetStatistic( TS, **kwargs ):

    """
    Calculate the Gaspard's test. 
    
    Parameters
    ----------
    TS : np.ndarray
        Sequence of symbols (or time series) to be analysed. Must be composed of discrete values, i.e. integers.
    n : int
        Length of the sequences of symbols to be analysed. Optional, default: 3.
    useShaders : bool
        Whether to use shaders or only CPU computation. Optional, default: False.

    Returns
    -------
    float
        Produced entropy.
    """


    return GetPValue( TS, numRndReps = 0, **kwargs )[ 1 ]





try:
    from numba import njit
    numba_installed = True
except ImportError:
    numba_installed = False

forceNoNumba = False

def optional_njit(*args, **kwargs):
    def decorator(func):
        if numba_installed and not forceNoNumba:
            return njit(*args, **kwargs)(func)
        else:
            return func
    return decorator






"""
Function to generate time series using a geometric Brownian motions with stochastic resetting (srGBM). Depending on the parameters, mainly the reset probability and the drift, resulting time series have a tuneable degree of irreversibility.

From:
    Stojkoski, V.; Sandev, T.; Kocarev, L.; Pal, A.
    Geometric Brownian motion under stochastic resetting: A stationary yet nonergodic process.
    Physical Review E 2021, 104, 014121.
"""

import numpy as np
from ..optionalNJIT import optional_njit
from ..warningHandling import warnUser


@optional_njit( cache=True, nogil=True )   
def srGBM( tsLen: int = 1000, drift: float = 0.05,
           noiseA: float = 0.01, r: float = 0.01 ):

    """
    Generate a time series using a geometric Brownian motion with stochastic resetting (srGBM).
    The srGBM is defined as:
    .. math:: x_{n+1} = x_n + \\mu x_n \Delta t + \\sigma x_n \\Delta W
    where :math:`x_n` is the value at time step :math:`n`, :math:`\\mu` is the drift, :math:`\\sigma` is the noise, and :math:`\\Delta W` is a Wiener process.
    The stochastic resetting is applied with a probability :math:`r` at each time step.
    
    Parameters
    ----------
    tsLen : int
        Length of the time series to generate.
    drift : float
        Drift of the geometric Brownian motion.
    noiseA : float
        Noise amplitude of the geometric Brownian motion.
    r : float
        Reset probability of the stochastic resetting.

    Returns
    -------
    np.ndarray
        Generated time series.
    """

    if tsLen < 1:
        raise ValueError( "tsLen must be greater than 0" )
    if r < 0.0 or r > 1.0:
        warnUser( "TimeSeriesGeneration.srGBM.srGBM", \
                  "r should be between 0.0 and 1.0" )

    TS = np.zeros( ( tsLen ) )
    TS[ 0 ] = 1.
    deltaT = 1.

    for t in range( 1, tsLen ):

        if np.random.uniform( 0., 1. ) < r * deltaT:
            TS[ t ] = 1.0
            continue

        TS[ t ] = TS[ t - 1 ]
        TS[ t ] += TS[ t - 1 ] * ( drift + np.sqrt( deltaT ) * noiseA * np.random.normal( 0.0, 1.0 ) )

    return TS





"""
Optimisation of the parameters for the Visibility Graph test

From:
    Lacasa, L., Nunez, A., Roldán, É., Parrondo, J. M., & Luque, B. (2012).
    Time series irreversibility: a visibility graph approach. The European Physical Journal B, 85(6), 1-11.
"""

import numpy as np
from itertools import product

from irreversibility.Metrics.VisibilityGraph import GetPValue


def Optimisation( tsSet, paramSet = None, criterion = np.median, numProcesses = -1 ):

    """
    Optimise the parameters for the Visibility Graph test.
    Note that this test has no parameters... hence the function is included for compatibility only.

    Parameters
    ----------
    tsSet : list of numpy.array
        Time series to be used in the optimisation.
    paramSet : list
        None required.
    criterion : function
        None required.
    numProcesses : int
        None required.
    kwargs : 
        None required.

    Returns
    -------
    dictionary
        Set of best parameters.
    float
        Lowest obtained p-value.
    """

    return {}, 1.0




"""
Ternary Coding test

From:
    Cammarota, C., & Rogora, E. (2007). 
    Time reversal, symbolic series and irreversibility of human heartbeat. 
    Chaos, Solitons & Fractals, 32(5), 1649-1654.
"""

import numpy as np
from statsmodels.stats.descriptivestats import sign_test
from ..optionalNJIT import optional_njit


    

@optional_njit( cache=True, nogil=True )             
def _numba_auxCountN( TS, alpha ):
    
    Np = 0
    Nn = 0
    
    TS = TS[ 1: ] - TS[ :-1 ]
    
    for k in range( TS.shape[0] - 2 ):
        if TS[ k ] > alpha and TS[ k + 1 ] > alpha and TS[ k + 2 ] > alpha:
            Np += 1
        if TS[ k ] < -alpha and TS[ k + 1 ] < -alpha and TS[ k + 2 ] < -alpha:
            Nn += 1
        
    return Np - Nn




def GetPValue( TS: np.array, **kwargs ):

    """
    Get p-value according to the Ternary Coding test.

    Parameters
    ----------
    TS : numpy.array
        Time series to be analysed.
    segL : int
        Length of the segments. Optional, default: 4.
    alpha : int
        Percentile used to evaluate if a value is extreme. Optional, default: 10.

    Returns
    -------
    float
        p-value of the test.
    float
        statistic of the test.

    Raises
    ------
    ValueError
        If the parameters are not correct.
    """


    # Setting up the parameters

    segL = kwargs.get( 'segL', 4 )
    alpha = kwargs.get( 'alpha', 10 )

    # Checking the parameters

    if type( segL ) is not int:
        raise ValueError("segL is not an integer")
    if segL <= 3:
        raise ValueError("segL must be larger than 3")
    if type( alpha ) is not int:
        raise ValueError("alpha is not an integer")
    if alpha <= 0:
        raise ValueError("alpha must be larger than 0")


    # Compute the test
        
    numSegments = int( np.floor( np.size( TS ) / segL ) )
    if numSegments < 3:
        return 1.0, 0.0
    
    D = TS[ 1: ] - TS[ :-1 ]
    alpha_star = np.percentile( D, alpha )
    
    Nd = []
    for k in range( numSegments ):
        tTS = TS[ ( k * segL ) : ( ( k + 1 ) * segL ) ]
        Nd.append( _numba_auxCountN( tTS, alpha_star ) )
    
    pValue, M = ( 1.0, 0.0 )
    try:
        M, pValue = sign_test( Nd )
    except:
        pass
    
    return pValue, np.abs( M )



def GetStatistic( TS: np.array, **kwargs ):

    """
    Get the statistic according to the Ternary Coding test.

    Parameters
    ----------
    TS : numpy.array
        Time series to be analysed.
    segL : int
        Length of the segments. Optional, default: 4.
    alpha : int
        Percentile used to evaluate if a value is extreme. Optional, default: 10.

    Returns
    -------
    float
        statistic of the test.

    Raises
    ------
    ValueError
        If the parameters are not correct.
    """

    return GetPValue( TS, **kwargs )[ 1 ]




"""
Skewness test

From:
    Koutsoyiannis, D. (2019)
    Time's arrow in stochastic characterization and simulation of atmospheric and hydrological processes.
    Hydrol. Sci. J. (64), 1013-1037.
    
    Vavoulogiannis, S., Iliopoulou, T., Dimitriadis, P., & Koutsoyiannis, D. (2021). 
    Multiscale temporal irreversibility of streamflow and its stochastic modelling. Hydrology, 8(2), 63.
"""

from scipy.stats import skew
from scipy.stats import norm
import numpy as np




def _calculateA( TS ):

    tdTS = TS[ 1: ] - TS[ :-1 ]
    skewness_orig = skew( TS )
    skewness_td = skew( tdTS )
    a = skewness_td / skewness_orig

    return a



def _A_For_Rnd_Series( TS, numRndReps = 100 ):

    a = []
    
    for k in range( numRndReps ):
        tTS = np.random.permutation( np.copy( TS ) )
        a.append( _calculateA( tTS ) )
    
    return a




def GetPValue( TS, **kwargs ):

    """
    Get p-value according to the Skewness test.

    Parameters
    ----------
    TS : numpy.array
        Time series to be analysed.
    numRndReps : int
        Number of random shuffled time series used to estimate the p-value. Optional, default: 100.
    pValueMethod : string
        Method used to estimate the p-value, including 'proportion' and 'z-score'. Optional, default: 'z-score'.

    Returns
    -------
    float
        p-value of the test.
    float
        statistic of the test.

    Raises
    ------
    ValueError
        If the parameters are not correct.
    """

    # Setting up the parameters

    numRndReps = kwargs.get( 'numRndReps', 100 )
    pValueMethod = kwargs.get( 'pValueMethod', 'z-score' )


    # Checking the parameters

    if type( numRndReps ) is not int:
        raise ValueError("numRndReps is not an integer")
    if numRndReps < 0:
        raise ValueError("numRndReps must be zero or positive")
    if pValueMethod not in [ 'proportion', 'z-score' ]:
        raise ValueError("pValueMethod not recognised")


    # Computing a for the original time series

    a = _calculateA( TS )

    if numRndReps <= 0:
        return 1.0, a


    # Extract the p-value

    pValue = 1.0
    rndA = _A_For_Rnd_Series( TS, numRndReps )

    if pValueMethod == 'z-score':
        z_score = ( a - np.mean( rndA ) ) / np.std( rndA )
        pValue = norm.sf( z_score )
    if pValueMethod == 'proportion':
        pValue = np.sum( np.abs( a ) < np.abs( np.array( rndA ) ) ) / numRndReps
    
    return pValue, a






def GetStatistic( TS, **kwargs ):

    """
    Get the statistic according to the Skewness test.

    Parameters
    ----------
    TS : numpy.array
        Time series to be analysed.

    Returns
    -------
    float
        statistic of the test.

    Raises
    ------
    ValueError
        If the parameters are not correct.
    """

    return GetPValue( TS, numRndReps = 0, **kwargs )[ 1 ]








import pytest
from pytest import approx

import numpy as np
import copy

import irreversibility as irr



@pytest.mark.parametrize("n", [2, 3, 4, 5])
def test_TS_Does_Not_Change( n ):

    for length in [ 100, 1000, 10000 ]:
        TS = np.random.randint( 0, 2+1, (length) )
        TS2 = copy.deepcopy( TS )
        currentS = np.random.randint( 0, 2+1, (n) )

        irr.DiscreteMetrics.Gaspard._numba_OneSequence( currentS, TS, n = n )
        assert np.sum( TS - TS2 ) == 0, (
            "Time series has been modified, numba"
        )

        irr.DiscreteMetrics.Gaspard._shader_OneSequence_Split( currentS, TS, n = n )
        assert np.sum( TS - TS2 ) == 0, (
            "Time series has been modified, shaders"
        )


@pytest.mark.parametrize("n", [2, 3, 4])
@pytest.mark.parametrize("length", [ 100, 1000, 10000 ])
def test_getPsi( n, length ):
    """Test the functions for calculating the entropy production"""

    TS = np.random.randint( 0, 2+1, (length) )
    e_1 = irr.DiscreteMetrics.Gaspard.EntropyProduction( TS, n = n, useShaders = False )
    e_2 = irr.DiscreteMetrics.Gaspard.EntropyProduction( TS, n = n, useShaders = True )
    assert e_1 == approx( e_2, rel=1e-5 ), (
        "The entropy production yielded by the standard approach and shaders' approach do not coincide (length: %d)" % length
    )


def test_GetPValue():
    """Test the functions for calculating the p-value"""

    TS = np.random.randint( 0, 2+1, (100) )

    with pytest.raises(ValueError):
        irr.DiscreteMetrics.Gaspard.GetPValue( TS, n = 1. )
    with pytest.raises(ValueError):
        irr.DiscreteMetrics.Gaspard.GetPValue( TS, n = -1. )

    with pytest.raises(ValueError):
        irr.DiscreteMetrics.Gaspard.GetPValue( TS, numRndReps = 10. )
    with pytest.raises(ValueError):
        irr.DiscreteMetrics.Gaspard.GetPValue( TS, numRndReps = -1 )

    with pytest.raises(ValueError):
        irr.DiscreteMetrics.Gaspard.GetPValue( TS, pValueMethod = 'a' )
    irr.DiscreteMetrics.Gaspard.GetPValue( TS, pValueMethod = 'proportion' )
    irr.DiscreteMetrics.Gaspard.GetPValue( TS, pValueMethod = 'z-score' )







"""
Two variants of a synthetic model to generate time series that
are irreversible at specific time scales.

An explanation can be found in:

Zanin, M., & Papo, D. (2025).
Algorithmic Approaches for Assessing Multiscale Irreversibility in Time Series: Review and Comparison.
Entropy, 27(2), 126.
"""

import numpy as np
from ..optionalNJIT import optional_njit
from ..warningHandling import warnUser


@optional_njit( cache=True, nogil=True )   
def msmodel( tsLen: int = 1000, tau: int = 2, mu: float = 0.05, \
             randomInit: bool = False ) -> np.ndarray :

    """
    Generate a time series using a multiscale model. See:
    Zanin, M., & Papo, D. (2025). Algorithmic Approaches for Assessing Multiscale Irreversibility in Time Series: Review and Comparison. Entropy, 27(2), 126.

    Parameters
    ----------
    tsLen : int
        Length of the time series to generate.
    tau : int
        Time scale of the model.
    mu : float
        Shift with respect to the previous value.
    randomInit : bool
        Not used, kept for compatibility with other functions.

    Returns
    -------
    np.ndarray
        Generated time series.
    """

    if mu < 0.0 or mu > 1.:
        warnUser( "TimeSeriesGeneration.MSModel.msmodel", \
                  "mu should be between 0.0 and 1.0" )

    if tsLen < 1:
        raise ValueError( "tsLen must be greater than 0" )
    if tau < 1:
        raise ValueError( "tau must be greater than 0" )

    TS = np.zeros( ( tsLen ) )
                                          
    for i in range( 0, tsLen ):

        if np.mod( i, tau ) == 0:
            TS[ i ] = np.random.uniform( 0., 1. )
        else:
            TS[ i ] = np.mod( TS[ i - 1 ] - mu, 1.0 )
    
    return TS


@optional_njit( cache=True, nogil=True )   
def msmodel_periodic( tsLen: int = 1000, tau: int = 2, \
                      mu: float = 0.05, randomInit: bool = False ) \
                      -> np.ndarray :

    """
    Generate a time series using a multiscale model with periodicity. See:
    Zanin, M., & Papo, D. (2025). Algorithmic Approaches for Assessing Multiscale Irreversibility in Time Series: Review and Comparison. Entropy, 27(2), 126.

    Parameters
    ----------
    tsLen : int
        Length of the time series to generate.
    tau : int
        Time scale of the model.
    mu : float
        Shift with respect to the previous value.
    randomInit : bool
        Not used, kept for compatibility with other functions.

    Returns
    -------
    np.ndarray
        Generated time series.
    """

    if mu < 0.0 or mu > 1.:
        warnUser( "TimeSeriesGeneration.MSModel.msmodel_periodic", \
                  "mu should be between 0.0 and 1.0" )

    if tsLen < 1:
        raise ValueError( "tsLen must be greater than 0" )
    if tau < 1:
        raise ValueError( "tau must be greater than 0" )

    TS = np.zeros( ( tsLen ) )
                                          
    for i in range( 0, tsLen ):

        if np.mod( i, tau + 1 ) < tau:
            TS[ i ] = np.random.uniform( 0., 1. )
        else:
            TS[ i ] = np.mod( TS[ i - tau ] - mu, 1.0 )
    
    return TS





"""
Local Clustering Coefficient test

From:
    Donges, J. F., Donner, R. V., & Kurths, J. (2013).
    Testing time series irreversibility using complex network methods.
    EPL (Europhysics Letters), 102(1), 10004.
"""

import numpy as np
from scipy.stats import ks_2samp
from ..optionalNJIT import optional_njit


    
        
@optional_njit( cache=True, nogil=True )            
def _numba_GetOneConnectivity( TS, offset ):
    
    tsLength = TS.shape[0]
    subAM = np.zeros( ( tsLength ) )
    if offset < tsLength - 1:
        subAM[ offset + 1 ] = 1

    if offset >= tsLength - 2: return subAM
    if TS[ offset + 1 ] >= TS[ offset ]: return subAM
        
    lastMax = TS[ offset + 1 ]

    for k in range( offset + 2, tsLength ):
        
        if TS[ k ] >= lastMax:
            lastMax = TS[ k ]
            subAM[ k ] = 1

        if lastMax > TS[ offset ]:
            break
            
    return subAM




def _GetFullAM( TS ):

    tsLength = np.size( TS, 0 )
    AM = np.zeros( (tsLength, tsLength) )
    
    for n1 in range( tsLength ):    
        AM[ n1, : ] = _numba_GetOneConnectivity( TS, n1 )
        
    for n1 in range( tsLength ):
        for n2 in range( n1 + 1, tsLength ):
            AM[ n2, n1 ] = AM[ n1, n2 ]
        
    return AM





@optional_njit( cache=True, nogil=True )            
def _numba_GetLocalClusteringCoefficient( AM ):
    
    numNodes = AM.shape[0]
    
    retardedCC = np.zeros( (numNodes) )
    for n1 in range( numNodes ):
        if np.sum( AM[ n1, :n1 ] ) <= 1.0: continue
        for n2 in range( n1 ):
            if AM[ n1, n2 ] == 0: continue
            for n3 in range( n1 ):
                 retardedCC[ n1 ] += AM[ n1, n2 ] * AM[ n2, n3 ] * AM[ n3, n1 ]
    
    advancedCC = np.zeros( (numNodes) )
    for n1 in range( numNodes ):
        if np.sum( AM[ n1, (n1+1): ] ) <= 1.0: continue
        for n2 in range( n1 + 1, numNodes ):
            if AM[ n1, n2 ] == 0: continue
            for n3 in range( n1 + 1, numNodes ):
                 advancedCC[ n1 ] += AM[ n1, n2 ] * AM[ n2, n3 ] * AM[ n3, n1 ]
                
    return retardedCC, advancedCC





    
def GetPValue( TS, **kwargs ):

    """
    Get p-value according to the Local Clustering Coefficient test.

    Parameters
    ----------
    TS : numpy.array
        Time series to be analysed.

    Returns
    -------
    float
        p-value of the test.
    float
        statistic of the test.

    Raises
    ------
    ValueError
        If the parameters are not correct.
    """

    AM = _GetFullAM( TS )
    retardedCC, advancedCC = _numba_GetLocalClusteringCoefficient( AM )
    stat, pV = ks_2samp( retardedCC,  advancedCC )

    return pV, stat



    
def GetStatistic( TS, **kwargs ):

    """
    Get the statistic of the Local Clustering Coefficient test.

    Parameters
    ----------
    TS : numpy.array
        Time series to be analysed.

    Returns
    -------
    float
        statistic of the test.

    Raises
    ------
    ValueError
        If the parameters are not correct.
    """

    return GetPValue( TS, **kwargs )[ 1 ]





"""
Pomeau test

From:
    Pomeau, Y. (1982). Symétrie des fluctuations dans le renversement du temps.
    Journal de Physique, 43(6), 859-867.
"""

import numpy as np
from scipy.stats import norm
from ..optionalNJIT import optional_njit




@optional_njit( cache=True, nogil=True )            
def _numba_getPsi( TS, tau = 1 ):
    
    psi = 0.0
    tsLen = TS.shape[0]

    for k in range( tsLen - 3 * tau - 1 ):
    
        psi += TS[ k ] * ( TS[ k + 2 * tau ] - TS[ k + tau ] ) * TS[ k + 3 * tau ]
    
    psi /= float( tsLen - 3 * tau - 1 )
    
    return psi
    



def _shader_getC_Split( TS, tau ):
    from wgpu.utils.compute import compute_with_buffers
    m = np.size( TS )

    A_shape = ( m )

    newTS = TS.astype(np.float32)

    bindings = {
        0: newTS,
        3: np.array(A_shape, dtype=np.uint32),
    }

    shader_src = """
    @group(0) @binding(0)
    var<storage, read> A: array<f32>;
    @group(0) @binding(2)
    var<storage, read_write> C: array<f32>;

    @group(0) @binding(3)
    var<storage, read> A_shape: array<u32>;

    override tau: u32;

    @compute @workgroup_size(1)
    fn main(@builtin(global_invocation_id) gid: vec3<u32>) {

        let m: u32 = A_shape[0];
        var sum: f32 = 0.0;

        var j: u32 = gid.x;
        sum = A[ j ] * ( A[ j + 2 * tau ] - A[ j + tau ] ) * A[ j + 3 * tau ];
        C[ j ] = sum;

        return;
    }
    """


    out = compute_with_buffers(
        input_arrays=bindings,
        output_arrays={2: (m, "f")},
        shader=shader_src,
        constants={'tau': tau},
        n=(m, 1, 1), 
    )

    C = np.frombuffer(out[2], dtype=np.float32)[ :( - 3 * tau - 1 )]
    return np.sum( C )


def _shader_getPsi( TS, tau ):
    C = 0
    if np.size( TS ) < 60000:
        C += _shader_getC_Split( TS, tau = tau )
    else:
        startOffset = 0
        while startOffset + 60000 + 3 * tau + 1 < np.size( TS ):
            C += _shader_getC_Split( TS[ startOffset : startOffset + 60000 + 3 * tau + 1 ], tau = tau )
            startOffset += 60000
        C += _shader_getC_Split( TS[ startOffset : ], tau = tau )

    psi = C / float( np.size( TS ) - 3 * tau - 1 )
    return psi



def GetPValue( TS: np.array, **kwargs ):

    """
    Get p-value according to the Pomeau test.

    Parameters
    ----------
    TS : numpy.array
        Time series to be analysed.
    tau : int
        Lag used in the computation of the function. Optional, default: 1.
    numRndReps : int
        Number of random shuffled time series used to estimate the p-value. Optional, default: 100.
    useShaders : bool
        Whether to use shaders or only CPU computation. Optional, default: False.
    pValueMethod : string
        Method used to estimate the p-value, including 'proportion' and 'z-score'. Optional, default: 'proportional'.

    Returns
    -------
    float
        p-value of the test.
    float
        statistic of the test.

    Raises
    ------
    ValueError
        If the parameters are not correct.
    """


    # Setting up the parameters

    tau = kwargs.get( 'tau', 1 )
    numRndReps = kwargs.get( 'numRndReps', 100 )
    useShaders = kwargs.get( 'useShaders', False )
    pValueMethod = kwargs.get( 'pValueMethod', 'proportion' )


    # Checking the parameters

    if type( tau ) is not int:
        raise ValueError("tau is not an integer")
    if tau <= 0:
        raise ValueError("tau must be larger than 0")
    if type( numRndReps ) is not int:
        raise ValueError("numRndReps is not an integer")
    if numRndReps < 0:
        raise ValueError("numRndReps must be zero or positive")
    if type( useShaders ) is not bool:
        raise ValueError("useShaders is not a bool")
    if pValueMethod not in [ 'proportion', 'z-score' ]:
        raise ValueError("pValueMethod not recognised")


    # Computing the psi for the original time series

    psi = 0.0

    if useShaders:
        psi = _shader_getPsi( TS, tau = tau )
    else:
        psi = _numba_getPsi( TS, tau = tau ) 
    
    if numRndReps <= 0:
        return 1.0, psi
    
    
    # Compute the psi for the shuffled versions of the time series

    rndPsi = []
    for k in range( numRndReps ):
        if useShaders:
            rndPsi.append( _shader_getPsi( np.random.permutation( TS ), tau = tau ) )
        else:
            rndPsi.append( _numba_getPsi( np.random.permutation( TS ), tau = tau ) )
    

    # Extract the p-value

    pValue = 1.0

    if pValueMethod == 'z-score':
        z_score = ( psi - np.mean( rndPsi ) ) / np.std( rndPsi )
        pValue = norm.sf( z_score )
    if pValueMethod == 'proportion':
        pValue = np.sum( np.abs( psi ) < np.abs( np.array( rndPsi ) ) ) / numRndReps
    
    return pValue, psi



def GetStatistic( TS, **kwargs ):

    """
    Get the statistic according to the Pomeau test.

    Parameters
    ----------
    TS : numpy.array
        Time series to be analysed.
    tau : int
        Lag used in the computation of the function. Optional, default: 1.
    useShaders : bool
        Whether to use shaders or only CPU computation. Optional, default: False.

    Returns
    -------
    float
        statistic of the test.

    Raises
    ------
    ValueError
        If the parameters are not correct.
    """

    return GetPValue( TS, numRndReps = 0, **kwargs )[ 1 ]





"""
Optimisation of the parameters for the Skewness test

From:
    Koutsoyiannis, D. (2019)
    Time's arrow in stochastic characterization and simulation of atmospheric and hydrological processes.
    Hydrol. Sci. J. (64), 1013-1037.
    
    Vavoulogiannis, S., Iliopoulou, T., Dimitriadis, P., & Koutsoyiannis, D. (2021). 
    Multiscale temporal irreversibility of streamflow and its stochastic modelling. Hydrology, 8(2), 63.
"""

import numpy as np
from itertools import product

from irreversibility.Metrics.Skewness import GetPValue


def Optimisation( tsSet, paramSet = None, criterion = np.median, numProcesses = -1 ):

    """
    Optimise the parameters for the Skewness test.
    Note that this test has no parameters... hence the function is included for compatibility only.

    Parameters
    ----------
    tsSet : list of numpy.array
        Time series to be used in the optimisation.
    paramSet : list
        None required.
    criterion : function
        None required.
    numProcesses : int
        None required.
    kwargs : 
        None required.

    Returns
    -------
    dictionary
        Set of best parameters.
    float
        Lowest obtained p-value.
    """

    return {}, 1.0




"""
DFK test

From:
    Daw, C. S., Finney, C. E. A., & Kennel, M. B. (2000).
    Symbolic approach for measuring temporal “irreversibility”. Physical Review E, 62(2), 1912.
"""

import numpy as np
from scipy.stats import chisquare
from ..optionalNJIT import optional_njit


    
    
def _GetSymbolicTimeSeries( TS, n ):
    
    numPoints = TS.shape[0]
    sortedTS = np.argsort( TS )
    thresholds_arg = np.array( [ sortedTS[ int( (k+1) * numPoints / n ) ] for k in range( n-1 ) ] )
    thresholds = TS[ thresholds_arg ]
    
    symbTS = np.zeros( ( numPoints ) )
    for offset in range( numPoints ):
        
        if TS[ offset ] < thresholds[0]:
            continue
        
        if TS[ offset ] >= thresholds[-1]:
            symbTS[ offset ] = n - 1
            
        for k in range( 0, n-2 ):
            if TS[ offset ] >= thresholds[ k ] and TS[ offset ] < thresholds[ k+1 ]:
                symbTS[ offset ] = k + 1
                break

    return symbTS



@optional_njit( cache=True, nogil=True )            
def _numba_GetWordsTimeSeries( symbTS, n, L ):
   
    numPoints = symbTS.shape[0]
    wordsTS = np.zeros( ( numPoints - L ) )
    for offset in range( numPoints - L ):
        
        currentW = 0
        for l in range( L ):
            currentW += symbTS[ offset + l ] * np.power( n, l )
        wordsTS[ offset ] = currentW
        
    return wordsTS





def GetPValue( TS, **kwargs ):

    """
    Get p-value according to the DFK test.

    Parameters
    ----------
    TS : numpy.array
        Time series to be analysed.
    n : int
        Number of symbols to be used. Optional, default: 3.
    L : int
        Number of symbols merged to create each word. Optional, default: 3.

    Returns
    -------
    float
        p-value of the test.
    float
        statistic of the test.

    Raises
    ------
    ValueError
        If the parameters are not correct.
    """
    

    # Setting up the parameters

    n = kwargs.get( 'n', 3 )
    L = kwargs.get( 'L', 3 )


    # Checking the parameters

    if type( n ) is not int:
        raise ValueError("n is not an integer")
    if n <= 1:
        raise ValueError("n must be larger than 1")
    if type( L ) is not int:
        raise ValueError("L is not an integer")
    if L <= 1:
        raise ValueError("L must be larger than 1")


    # Computing the test

    numWords = int( np.power( n, L ) )
    
    if np.size( TS ) <= numWords * 2:
        return 1.0, 0.0
    
    symbTS = _GetSymbolicTimeSeries( TS, n )
    wordsTS = _numba_GetWordsTimeSeries( symbTS, n, L )
    freqWords = np.array( [ np.sum( wordsTS == word ) for word in range( numWords ) ], dtype = float )
    
    symbTS_r = _GetSymbolicTimeSeries( TS[::-1], n )
    wordsTS_r = _numba_GetWordsTimeSeries( symbTS_r, n, L )
    freqWords_r = np.array( [ np.sum( wordsTS_r == word ) for word in range( numWords ) ], dtype = float )
    
    if np.sum( freqWords < 0.5 ) + np.sum( freqWords_r < 0.5 ) > 0:
        freqWords += 0.001
        freqWords_r += 0.001
    
    chisq, pV = chisquare( freqWords, freqWords_r )
    
    if pV == 0.0:
        return 1.0, 0.0
    
    return pV, chisq



def GetStatistic( TS, **kwargs ):

    """
    Get the statistic according to the DFK test.

    Parameters
    ----------
    TS : numpy.array
        Time series to be analysed.
    n : int
        Number of symbols to be used. Optional, default: 3.
    L : int
        Number of symbols merged to create each word. Optional, default: 3.

    Returns
    -------
    float
        statistic of the test.

    Raises
    ------
    ValueError
        If the parameters are not correct.
    """

    return GetPValue( TS, **kwargs )[ 1 ]



"""
Simple function to generate a time series using the Logistic map.
"""

import numpy as np
from ..optionalNJIT import optional_njit
from ..warningHandling import warnUser


@optional_njit( cache=True, nogil=True )   
def logistic( tsLen: int = 1000, skip: int = 100, x: float = 0.01, \
              r: float = 4.0, randomInit: bool = False) -> np.ndarray :
    
    """
    Generate a time series using the Logistic map.
    The Logistic map is defined as:
    .. math:: x_{n+1} = r * x_n * (1 - x_n)
    where :math:`x_n` is the value at time step :math:`n`, and :math:`r` is a parameter.

    Parameters
    ----------
    tsLen : int
        Length of the time series to generate.
    skip : int
        Number of initial iterations to skip.
    x : float
        Initial value of the time series.
    r : float
        Parameter of the Logistic map.
    randomInit : bool
        If True, a random initial value is generated between 0.01 and 0.99.
        
    Returns
    -------
    np.ndarray
        Generated time series.
    """

    if randomInit:
        x = np.random.uniform( 0.01, 0.99 )

    if x < 0.0 or x > 1.:
        warnUser( "TimeSeriesGeneration.Logistic.logistic", \
                  "x should be between 0.0 and 1.0" )
    if r < 0.0 or r > 4.:
        warnUser( "TimeSeriesGeneration.Logistic.logistic", \
                  "r should be between 0.0 and 4.0" )
        
    if tsLen < 1:
        raise ValueError( "tsLen must be greater than 0" )
    if skip < 0:
        raise ValueError( "skip must be greater than or equal to 0" )

    for t in range( skip ):
        x = r * x * ( 1.0 - x )

    TS = np.zeros( (tsLen) )
                                          
    for t in range( tsLen ):
        x = r * x * ( 1.0 - x )
        TS[t] = x

    return TS




"""
Simple function to generate a time series using the Henon map.
"""

import numpy as np
from ..optionalNJIT import optional_njit
from ..warningHandling import warnUser


@optional_njit( cache=True, nogil=True )   
def henon( tsLen: int = 1000, x: float = 0.5, y: float = 0.5, a: float = 1.4, b: float = 0.3, randomInit: bool = False ) -> np.ndarray :

    """
    Generate a time series using the Henon map.
    The Henon map is defined as:
    .. math:: x_{n+1} = 1 - a * x_n^2 + y_n
    .. math:: y_{n+1} = b * x_n
    where :math:`x_n` and :math:`y_n` are the values at time step :math:`n`, and :math:`a` and :math:`b` are parameters.
    
    Parameters
    ----------
    tsLen : int
        Length of the time series to generate.
    x : float
        Initial value of x.
    y : float
        Initial value of y.
    a : float
        Parameter a of the Henon map.
    b : float
        Parameter b of the Henon map.
    randomInit : bool
        If True, a random initial value is generated for x and y.

    Returns
    -------
    np.ndarray
        Generated time series.
    """

    if randomInit:
        x = np.random.uniform( -1., 1. )
        y = np.random.uniform( -0.3, 0.3 )

    if x < -1.5 or x > 1.5:
        warnUser( "TimeSeriesGeneration.Henon.henon", \
                  "x should be between -1.5 and 1.5" )
    if y < -0.5 or y > 0.5:
        warnUser( "TimeSeriesGeneration.Henon.henon", \
                  "y should be between -0.5 and 0.5" )

    if tsLen < 1:
        raise ValueError( "tsLen must be greater than 0" )

    TS = np.zeros( (tsLen, 2) )
    TS[0, 0] = x
    TS[0, 1] = y
                                          
    for i in range(1, tsLen):

        TS[i, 0] = 1.0 - a * TS[i - 1, 0] * TS[i - 1, 0] + TS[i - 1, 1]
        TS[i, 1] = b * TS[i - 1, 0]
    
    return TS





"""
Optimisation of the parameters for the Costa index test

From:
    Costa, M.; Goldberger, A.L.; Peng, C.K.
    Broken asymmetry of the human heartbeat: loss of time irreversibility in aging and disease.
    Physical review letters 2005, 95, 198102.
"""

import numpy as np
from itertools import product
from multiprocessing import Pool
from functools import partial

from irreversibility.Metrics.CostaIndex import GetPValue




def Optimisation( tsSet, paramSet = None, criterion = np.median, numProcesses = -1, **kwargs ):

    """
    Optimise the parameters for the Costa index test.

    Parameters
    ----------
    tsSet : list of numpy.array
        Time series to be used in the optimisation.
    paramSet : list
        List of parameters' values to be evaluated. For each parameter (in the case of this test, one), the list must contain a list of possible values, e.g. [ [1, 2] ]. If None, a standard set is used. Optional, default: None.
    criterion : function
        Function to be applied to the set of p-values to obtain the best option. Optional, default: numpy.median.
    numProcesses : int
        Number of parallel tasks used in the evaluation. If -1, only one task is used. Optional, default: False.
    kwargs : 
        Other options to be passed to the function to obtain the p-values.

    Returns
    -------
    dictionary
        Set of best parameters.
    float
        Lowest obtained p-value.

    Raises
    ------
    ValueError
        If the parameters are not correct.
    """

    if paramSet is None:
        p1Set = [ 1, 2, 3 ]
    else:
        p1Set = paramSet[ 0 ]

    bestPValue = 1.0
    bestParameters = []

    for pSet in p1Set:

        pV = np.zeros( ( len( tsSet ) ) )

        if numProcesses == -1:

            for k in range( len( tsSet ) ):
                pV[ k ] = GetPValue( tsSet[ k ], tau = pSet, **kwargs )[ 0 ]

        else:

            with Pool( processes = numProcesses ) as pool:

                async_result = []
                for k in range( len( tsSet ) ):
                    func = partial( GetPValue, TS = tsSet[ k ], tau = pSet, **kwargs )
                    async_result.append( pool.apply_async( func ) )
                
                [result.wait() for result in async_result]
                for k in range( len( tsSet ) ):
                    pV[ k ] = async_result[ k ].get()[0]

        synthPV = criterion( pV )
        if synthPV < bestPValue:
            bestPValue = synthPV
            bestParameters = { 'tau': pSet }

    return bestParameters, bestPValue




import pytest
from pytest import approx

import numpy as np
import copy

import irreversibility as irr



@pytest.mark.parametrize("tau", [1, 2, 3, 4, 5])
def test_TS_Does_Not_Change( tau ):

    for length in [ 100, 1000, 10000 ]:
        TS = np.random.uniform( 0., 1., ( length ) )
        TS2 = copy.deepcopy( TS )

        irr.Metrics.Pomeau._numba_getPsi( TS, tau = tau )
        assert np.sum( TS - TS2 ) == 0, (
            "Time series has been modified, numba"
        )

        irr.Metrics.Pomeau._shader_getPsi( TS, tau = tau )
        assert np.sum( TS - TS2 ) == 0, (
            "Time series has been modified, shaders"
        )


@pytest.mark.parametrize("tau", [1, 2, 3])
def test_getPsi( tau ):
    """Test the functions for calculating Psi"""

    TS = np.array( range( 100 ) ) / 100.
    correctPsi = [ 
        0.,
        0.003166666,
        0.006225333,
        0.009166999
    ]

    psi_1 = irr.Metrics.Pomeau._numba_getPsi( TS, tau = tau )
    assert psi_1 == approx( correctPsi[ tau ] ), (
        "Psi yielded by the standard approach is not correct"
    )

    psi_2 = irr.Metrics.Pomeau._shader_getPsi( TS, tau = tau )
    assert psi_1 == approx( psi_2 ), (
        "Psi yielded by the standard approach and shaders' approach do not coincide"
    )

    for length in [ 100, 1000, 10000 ]:
        TS = np.random.uniform( 0., 1., ( length ) )
        psi_1 = irr.Metrics.Pomeau._numba_getPsi( TS, tau = tau )
        psi_2 = irr.Metrics.Pomeau._shader_getPsi( TS, tau = tau )
        assert psi_1 == approx( psi_2, rel=1e-3 ), (
            "Psi yielded by the standard approach and shaders' approach do not coincide (length: %d)" % length
        )


@pytest.mark.parametrize("length", [ 100, 1000, 10000 ])
def test_GetPValue( length ):
    """Test the functions for calculating the p-value"""

    TS = np.random.uniform( 0., 1., (length) )

    with pytest.raises(ValueError):
        irr.Metrics.Pomeau.GetPValue( TS, tau = 1. )
    with pytest.raises(ValueError):
        irr.Metrics.Pomeau.GetPValue( TS, tau = -1. )

    with pytest.raises(ValueError):
        irr.Metrics.Pomeau.GetPValue( TS, numRndReps = 10. )
    with pytest.raises(ValueError):
        irr.Metrics.Pomeau.GetPValue( TS, numRndReps = -1 )

    with pytest.raises(ValueError):
        irr.Metrics.Pomeau.GetPValue( TS, pValueMethod = 'a' )
    irr.Metrics.Pomeau.GetPValue( TS, pValueMethod = 'proportion' )
    irr.Metrics.Pomeau.GetPValue( TS, pValueMethod = 'z-score' )



@pytest.mark.parametrize("length", [ 100, 1000, 10000 ])
def test_GetStatistic( length ):
    """Test the functions for calculating the p-value"""

    TS = np.random.uniform( 0., 1., (length) )

    result = irr.Metrics.Pomeau.GetStatistic( TS )
    assert isinstance( result, float )



def test_GetStatistic():
    """Test the functions for calculating Psi"""

    TS = np.random.uniform( 0., 1., ( 100 ) )

    with pytest.raises(ValueError):
        irr.Metrics.Pomeau.GetStatistic( TS, tau = 1. )
    with pytest.raises(ValueError):
        irr.Metrics.Pomeau.GetStatistic( TS, tau = -1. )

    with pytest.raises(ValueError):
        irr.Metrics.Pomeau.GetPValue( TS, numRndReps = 10. )
    with pytest.raises(ValueError):
        irr.Metrics.Pomeau.GetPValue( TS, numRndReps = -1 )

    with pytest.raises(ValueError):
        irr.Metrics.Pomeau.GetPValue( TS, pValueMethod = 'a' )
    irr.Metrics.Pomeau.GetPValue( TS, pValueMethod = 'proportion' )
    irr.Metrics.Pomeau.GetPValue( TS, pValueMethod = 'z-score' )


def test_Optimisation():
    """Test the functions for optimising the parameters"""

    from irreversibility.ParameterOptimisation.Opt_Pomeau import Optimisation

    tsSet = []
    for k in range( 10 ):
        tsSet.append( np.random.uniform( 0., 1., (6453) ) )

    Optimisation( tsSet, paramSet = None, criterion = np.median, numProcesses = -1, useShaders = False )
    Optimisation( tsSet, paramSet = None, criterion = np.median, numProcesses = -1, useShaders = True )





"""
Ramsey test

From:
    Ramsey, J. B., & Rothman, P. (1996).
    Time irreversibility and business cycle asymmetry. Journal of Money, Credit and Banking, 28(1), 1-21.
"""

import numpy as np
from scipy.stats import ttest_ind
from ..optionalNJIT import optional_njit



@optional_njit( cache=True, nogil=True )            
def _numba_auxFunction( TS, kappa ):
    
    numPoints = TS.shape[0]

    mean1 = np.zeros( ( numPoints - 2 * kappa ) )
    mean2 = np.zeros( ( numPoints - 2 * kappa ) )
    for k in range( kappa, numPoints - kappa ):
        mean1[ k - kappa ] = TS[k] * TS[k] * TS[k - kappa]
        mean2[ k - kappa ] = TS[k] * TS[k - kappa] * TS[k - kappa]

    return mean1, mean2    




def GetPValue( TS, **kwargs ):

    """
    Get p-value according to the Ramsey test.

    Parameters
    ----------
    TS : numpy.array
        Time series to be analysed.
    kappa : int
        Lag used in the computation of the function. Optional, default: 1.

    Returns
    -------
    float
        p-value of the test.
    float
        statistic of the test.

    Raises
    ------
    ValueError
        If the parameters are not correct.
    """

    # Setting up the parameters

    kappa = kwargs.get( 'kappa', 1 )


    # Checking the parameters

    if type( kappa ) is not int:
        raise ValueError("kappa is not an integer")
    if kappa <= 0:
        raise ValueError("kappa must be larger than 0")


    # Computing the test

    TS = TS - np.mean( TS )

    mean1, mean2 = _numba_auxFunction( TS, kappa )
    
    _, pV = ttest_ind( mean1, mean2 )
    
    return pV, np.mean( mean2 ) - np.mean( mean1 )



def GetStatistic( TS, **kwargs ):

    """
    Get the statistic according to the Ramsey test.

    Parameters
    ----------
    TS : numpy.array
        Time series to be analysed.
    kappa : int
        Lag used in the computation of the function. Optional, default: 1.

    Returns
    -------
    float
        statistic of the test.

    Raises
    ------
    ValueError
        If the parameters are not correct.
    """

    return GetPValue( TS, **kwargs )[ 1 ]




import pytest
from pytest import approx

import numpy as np
import copy

import irreversibility as irr



@pytest.mark.parametrize("pSize", [3, 4, 5])
@pytest.mark.parametrize("length", [ 100, 1000, 10000 ])
def test_TS_Does_Not_Change( pSize, length ):

    x0 = np.random.uniform( 0.0, 1.0, (pSize) )
    TS = np.random.uniform( 0., 1., ( length ) )
    rTS = TS[::-1]
    TS2 = copy.deepcopy( TS )

    irr.Metrics.COP._evaluateDistance( x0, TS, rTS, useShaders = False )
    assert np.sum( TS - TS2 ) == 0, (
        "Time series has been modified, numba"
    )

    irr.Metrics.COP._evaluateDistance( x0, TS, rTS, useShaders = True )
    assert np.sum( TS - TS2 ) == 0, (
        "Time series has been modified, shaders"
    )


@pytest.mark.parametrize("pSize", [3, 4, 5])
@pytest.mark.parametrize("length", [ 100, 1000, 10000, 100000 ])
def test_tranfTS( pSize, length ):
    """Test the functions for calculating Psi"""

    TS = np.random.uniform( 0., 1., ( length ) )
    patt = np.random.uniform( 0., 1., ( pSize ) )
    patt = irr.Metrics.COP._numba_normWindow( patt )

    psi_1 = irr.Metrics.COP._shader_tranfTS( TS, patt )
    psi_2 = irr.Metrics.COP._numba_tranfTS( TS, patt )
    assert np.sum( psi_1 ) == approx( np.sum( psi_2 ), rel=1e-3 ), (
        "Distance yielded by the standard approach and shaders' approach do not coincide (length: %d)" % length
    )


@pytest.mark.parametrize("length", [ 100, 1000, 10000 ])
def test_GetPValue( length ):
    """Test the functions for calculating the p-value"""

    TS = np.random.uniform( 0., 1., (length) )

    with pytest.raises(ValueError):
        irr.Metrics.COP.GetPValue( TS, pSize = 1. )
    with pytest.raises(ValueError):
        irr.Metrics.COP.GetPValue( TS, pSize = -1 )

    with pytest.raises(ValueError):
        irr.Metrics.COP.GetPValue( TS, numIters = 10. )
    with pytest.raises(ValueError):
        irr.Metrics.COP.GetPValue( TS, numIters = -1 )

    irr.Metrics.COP.GetPValue( TS )



@pytest.mark.parametrize("length", [ 100, 1000, 10000 ])
def test_GetStatistic( length ):
    """Test the functions for calculating the p-value"""

    TS = np.random.uniform( 0., 1., (length) )

    result = irr.Metrics.COP.GetStatistic( TS )
    assert isinstance( result, float )




def test_Optimisation():
    """Test the functions for optimising the parameters"""

    from irreversibility.ParameterOptimisation.Opt_COP import Optimisation

    tsSet = []
    for k in range( 10 ):
        tsSet.append( np.random.uniform( 0., 1., (645) ) )

    Optimisation( tsSet, paramSet = None, criterion = np.median, numProcesses = -1, useShaders = False )
    Optimisation( tsSet, paramSet = None, criterion = np.median, numProcesses = -1, useShaders = True )





"""
Casali test.

From:
    Casali, K. R., Casali, A. G., Montano, N., Irigoyen, M. C., Macagnan, F., Guzzetti, S., & Porta, A. (2008).
    Multiple testing strategy for the detection of temporal irreversibility in stationary time series.
    Physical Review E—Statistical, Nonlinear, and Soft Matter Physics, 77(6), 066204.
"""

import numpy as np
from scipy.stats import norm
from ..optionalNJIT import optional_njit




@optional_njit( cache=True, nogil=True )            
def _numba_CalculateS( TS, m ):

    k = 2. ** (- 0.5 )
    N = TS.shape[ 0 ]

    S_m = 0.
    denom = 0.
    for i in range( 1, N - m ):
        d_mi = k * ( TS[ i ] - TS[ i + m ] )
        S_m += d_mi ** 3.
        denom += d_mi ** 2.

    S_m /= denom ** ( 3. / 2. )

    return S_m



def GetPValue( TS, **kwargs ):

    """
    Get p-value according to the Casali test.

    Parameters
    ----------
    TS : numpy.array
        Time series to be analysed.
    m : int
        Lag used in the computation of the function. Optional, default: 1.
    numRndReps : int
        Number of random shuffled time series used to estimate the p-value. Optional, default: 100.
    pValueMethod : string
        Method used to estimate the p-value, including 'proportion' and 'z-score'. Optional, default: 'proportional'.

    Returns
    -------
    float
        p-value of the test.
    float
        statistic of the test.

    Raises
    ------
    ValueError
        If the parameters are not correct.
    """

    # Setting up the parameters

    m = kwargs.get( 'm', 1 )
    numRndReps = kwargs.get( 'numRndReps', 100 )
    pValueMethod = kwargs.get( 'pValueMethod', 'proportion' )


    # Checking the parameters

    if type( m ) is not int:
        raise ValueError("m is not an integer")
    if m <= 0:
        raise ValueError("m must be larger than 0")
    if type( numRndReps ) is not int:
        raise ValueError("numRndReps is not an integer")
    if numRndReps < 0:
        raise ValueError("numRndReps must be zero or positive")
    if pValueMethod not in [ 'proportion', 'z-score' ]:
        raise ValueError("pValueMethod not recognised")


    # Computing the statistics for the original time series

    S = _numba_CalculateS( TS, m = m )
    
    if numRndReps <= 0:
        return 1.0, S
    
    
    # Compute the statistics for the shuffled versions of the time series

    rndS = []
    for k in range( numRndReps ):
        rndS.append( _numba_CalculateS( np.random.permutation( TS ), m = m ) )


    # Extract the p-value

    pValue = 1.0

    if pValueMethod == 'z-score':
        z_score = ( S - np.mean( rndS ) ) / np.std( rndS )
        pValue = norm.sf( z_score )
    if pValueMethod == 'proportion':
        pValue = np.sum( np.abs( S ) < np.abs( np.array( rndS ) ) ) / numRndReps
    
    return pValue, S




def GetStatistic( TS, **kwargs ):

    """
    Get the statistics according to the Casali test.

    Parameters
    ----------
    TS : numpy.array
        Time series to be analysed.
    m : int
        Lag used in the computation of the function. Optional, default: 1.

    Returns
    -------
    float
        statistic of the test.

    Raises
    ------
    ValueError
        If the parameters are not correct.
    """

    return GetPValue( TS, numRndReps = 0, **kwargs )[ 1 ]







import numpy as np
from ..optionalNJIT import optional_njit
from ..warningHandling import warnUser


@optional_njit( cache=True, nogil=True )   
def lorenz( tsLen: int = 1000, dT: float = 0.01, x: float = 0.0, \
            y: float = 1.0, z: float = 1.05, \
            s: float = 10, r: float = 28, \
            b: float = 2.667, randomInit: bool = False ) -> np.ndarray :
    
    """
    Generate a time series using the Lorenz system.
    The Lorenz system is defined by the following equations:
    .. math:: \\frac{dx}{dt} = s (y - x)
    .. math:: \\frac{dy}{dt} = r x - y - x z
    .. math:: \\frac{dz}{dt} = x y - b z
    where :math:`x`, :math:`y`, and :math:`z` are the variables of the system,
    and :math:`s`, :math:`r`, and :math:`b` are parameters.

    Parameters
    ----------
    tsLen : int
        Length of the time series to generate.
    dT : float
        Time step for the simulation.
    x : float
        Initial value of the x variable.
    y : float
        Initial value of the y variable.
    z : float
        Initial value of the z variable.
    s : float
        Parameter of the Lorenz system.
    r : float
        Parameter of the Lorenz system.
    b : float
        Parameter of the Lorenz system.
    randomInit : bool
        If True, random initial values are generated for x, y, and z.
    
    Returns
    -------
    np.ndarray
        Generated time series.
    """
      
    if randomInit:
        x = np.random.uniform( -0.5, 0.5 )
        y = np.random.uniform( 0.0, 2.0 )
        z = np.random.uniform( 0.9, 1.1 )

    if s < 0.0:
        warnUser( "TimeSeriesGeneration.Lorenz.lorenz", \
                  "s should be positive" )
    if r < 0.0:
        warnUser( "TimeSeriesGeneration.Lorenz.lorenz", \
                  "r should be positive" )
    if b < 0.0:
        warnUser( "TimeSeriesGeneration.Lorenz.lorenz", \
                  "b should be positive" )
        
    if tsLen < 1:
        raise ValueError( "tsLen must be greater than 0" )
    if dT <= 0.0:
        raise ValueError( "dT must be greater than 0" )
    
    tsLen = int( tsLen )
    TS = np.zeros( (tsLen, 3) )
    TS[0, 0] = x
    TS[0, 1] = y
    TS[0, 2] = z
                                          
    for i in range(1, tsLen):

        x_dot = s * ( TS[i - 1, 1] - TS[i - 1, 0] )
        y_dot = r * TS[i - 1, 0] - TS[i - 1, 1] - TS[i - 1, 0] * TS[i - 1, 2]
        z_dot = TS[i - 1, 0] * TS[i - 1, 1] - b * TS[i - 1, 2]

        TS[i, 0] = TS[i - 1, 0] + (x_dot * dT)
        TS[i, 1] = TS[i - 1, 1] + (y_dot * dT)
        TS[i, 2] = TS[i - 1, 2] + (z_dot * dT)
    
    return TS






from irreversibility.Metrics.BDS import GetPValue as BDSTest
from irreversibility.Metrics.Casali import GetPValue as CasaliTest
from irreversibility.Metrics.COP import GetPValue as COPTest
from irreversibility.Metrics.CostaIndex import GetPValue as CostaTest
from irreversibility.Metrics.DFK import GetPValue as DFKTest
from irreversibility.Metrics.Diks import GetPValue as DiksTest
from irreversibility.Metrics.LocalCC import GetPValue as LocalCCTest
from irreversibility.Metrics.MSTrends import GetPValue as MSTrendsTest
from irreversibility.Metrics.PermPatterns import GetPValue as PPTest
from irreversibility.Metrics.Pomeau import GetPValue as PomeauTest
from irreversibility.Metrics.Ramsey import GetPValue as RamseyTest
from irreversibility.Metrics.Skewness import GetPValue as SkewnessTest
from irreversibility.Metrics.TernaryCoding import GetPValue as TCTest
from irreversibility.Metrics.TPLength import GetPValue as TPTest
from irreversibility.Metrics.VisibilityGraph import GetPValue as VGTest
from irreversibility.Metrics.Zumbach import GetPValue as ZumbachTest

allTests = [
    [ 'BDS', BDSTest ],
    [ 'Casali', CasaliTest ],
    [ 'COP', COPTest ],
    [ 'Costa Index', CostaTest ],
    [ 'DFK', DFKTest ],
    [ 'Diks', DiksTest ],
    [ 'Local CC', LocalCCTest ],
    [ 'MS Trends', MSTrendsTest ],
    [ 'Permutation patterns', PPTest ],
    [ 'Pomeau', PomeauTest ],
    [ 'Ramsey', RamseyTest ],
    [ 'Skewness', SkewnessTest ],
    [ 'Ternary Coding', TCTest ],
    [ 'TP Length', TPTest ],
    [ 'Visibility Graph', VGTest ],
    [ 'Zumbach', ZumbachTest ],
]




"""
Zumbach test

From:
    Zumbach, G., (2007).
    Time reversal invariance in finance.
    https://arxiv.org/pdf/0708.4022.pdf
"""

import numpy as np
from scipy.stats import skew
from scipy.stats import norm




def _calcAllReturns( ts, granularity ):

    ts = np.log2( ts )
    return ts[ granularity : ] - ts[ : (-granularity) ]



def _histVolatility( ts, t, deltaT, allRets ):

    ax = []
    for k in range( t - deltaT, t ):
        if k < 0: continue
        if k >= np.size( allRets ): continue
        ax.append( allRets[ k ] ** 2.0 )
    
    return np.mean( ax )



def _realVolatility( ts, t, deltaT, allRets, granularity ):

    ax = []
    for k in range( t + granularity, t + deltaT ):
        if k >= np.size( allRets ): continue
        ax.append( allRets[ k ] ** 2.0 )
    
    return np.mean( ax )





def GetPValue( TS, **kwargs ):

    """
    Get p-value according to the Zumbach test.

    Parameters
    ----------
    TS : numpy.array
        Time series to be analysed.
    deltaT : int
        Size of the window used to calculate the volatility. Optional, default: 20.
    granularity : int
        Number of values skipped to calculate the real volatility. Optional, default: 3.

    Returns
    -------
    float
        p-value of the test.
    float
        statistic of the test.

    Raises
    ------
    ValueError
        If the parameters are not correct.
    """

    # Setting up the parameters

    deltaT = kwargs.get( 'deltaT', 20 )
    granularity = kwargs.get( 'granularity', 3 )


    # Checking the parameters

    if type( deltaT ) is not int:
        raise ValueError("deltaT is not an integer")
    if deltaT <= 1:
        raise ValueError("deltaT must be larger than one")
    if type( granularity ) is not int:
        raise ValueError("granularity is not an integer")
    if granularity <= 1:
        raise ValueError("granularity must be larger than one")


    # Calculate the test

    allRets = _calcAllReturns( TS, granularity )

    set1 = []
    set2 = []
    for t in range( np.size( TS ) ):
        set1.append( _histVolatility( TS, t, deltaT, allRets ) )
        set2.append( _realVolatility( TS, t, deltaT, allRets, granularity ) )

    set1 = np.array( set1 )
    set2 = np.array( set2 )
    notValid = np.any( ( np.isnan( set1 ), np.isnan( set2 ) ), axis = 0 )

    set1 = set1[ ~notValid ]
    set2 = set2[ ~notValid ]

    from scipy.stats import ks_2samp
    if np.size( set1 ) == 0 or np.size( set2 ) == 0:
        return 1.0, 0.0
    
    stat, pV = ks_2samp( set1,  set2 )

    return pV, stat



def GetStatistic( TS, **kwargs ):

    """
    Get the statistic according to the Zumbach test.

    Parameters
    ----------
    TS : numpy.array
        Time series to be analysed.
    deltaT : int
        Size of the window used to calculate the volatility. Optional, default: 20.
    granularity : int
        Number of values skipped to calculate the real volatility. Optional, default: 3.

    Returns
    -------
    float
        statistic of the test.

    Raises
    ------
    ValueError
        If the parameters are not correct.
    """

    return GetPValue( TS, **kwargs )[ 1 ]





"""
Function to manipulate time series using Continuous Ordinal Patterns, 
aimed at increasing (or decreasing) their irreversibility.
For the theory behind this, see:

Zanin, M. (2024).
Manipulating Time Series Irreversibility Through Continuous Ordinal Patterns.
Symmetry, 16(12), 1696.
"""

import numpy as np

from irreversibility.Metrics.COP import _numba_normWindow, _numba_tranfTS

from ..warningHandling import warnUser



def GetTransformedTimeSeries( timeSeries: np.ndarray, test, testParams, \
                              increase: bool = True, pSize: int = 4, \
                              numIterations: int = 1000, \
                              pvThreshold: float = 0.001 ):

    """
    Try to transform the input time series using Continuous Ordinal Patterns, to either increase or decrease its irreversibility. The manipulation is performed by trying to minimise the magnitude of the changes, as measured by a linear correlation.
    
    Parameters
    ----------
    timeSeries : np.ndarray
        Input time series to be manipulated.
    test : function
        Function to be used to calculate the irreversibility. Must be one of the 'GetPValue' functions of this library.
    testParams : dictionary
        Additional parameters to be passed to the irreversibility test function.
    increase : bool
        Defines whether the aim is to increase (true) or decrease (false) the irreversibility
    pSize : int
        Length of the COP used in the transformation.
    numIterations : int
        Number of times a different random transformation is tested.
    pvThreshold : float
        Threshold used to assess if a solution is accepted. When the target is to increase the irreversibility, the p-value of the transformed time series must be below this threshold; the opposite in the case of reducing the irreversibility.

    Returns
    -------
    np.ndarray
        Manipulated time series. If no acceptable solution is found, this corresponds to the original time series.
    float
        Best p-value obtained, i.e. the one of the manipulated time series.
    float
        Linear correlation between the original and manipulated time series.
    """

    if pSize < 2:
        raise ValueError( "pSize cannot be smaller than 2" )
    if pSize == 2:
        warnUser( "Manipulation.COP.GetTransformedTimeSeries", \
                  "While pSize can technically be 2, it is recommended to use larger values." )
    if numIterations < 1:
        raise ValueError( "numIterations cannot be smaller than 1" )

    if increase:
        bestPV = 1.0
    else:
        bestPV = 0.0
        
    bestCorr = 0.0
    bestTS = []

    for k in range( numIterations ):

        x0 = np.random.uniform( 0.0, 1.0, (pSize) )
        patt = _numba_normWindow( x0 )
        w = _numba_tranfTS( timeSeries, patt )

        corrc = np.abs( np.corrcoef( [ timeSeries[ :np.size( w ) ], w ] )[ 0, 1 ] )
        p_value, _ = test( w, **testParams )

        if increase and p_value >= pvThreshold: continue
        if ( not increase ) and p_value < pvThreshold: continue
        if corrc > bestCorr:
            bestPV = p_value
            bestCorr = corrc
            bestTS = w

    if type( bestTS ) == list:
        return bestTS, bestPV, bestCorr
    
    if np.corrcoef( [ timeSeries[ :np.size( bestTS ) ], bestTS ] )[ 0, 1 ] < 0.0:
        bestTS = 1.0 - bestTS

    return bestTS, bestPV, bestCorr





"""
Diks test

From:
    Diks, C., Van Zwet, W. R., Takens, F., & DeGoede, J. (1996).
    Detecting differences between delay vector distributions. Physical Review E, 53(3), 2169.
"""

import numpy as np
from scipy.special import perm
import scipy.stats as st
from ..optionalNJIT import optional_njit




@optional_njit( cache=True, nogil=True )
def _numba_CreateVectors( TS, embD ):
    
    dimX = int( ( TS.shape[0] - embD ) / embD )
    vectors = np.zeros( ( dimX, embD ) )
    for k in range( 0, dimX ):
        offset = k * embD
        vectors[ k, : ] = TS[ offset : (offset + embD) ]

    return vectors





@optional_njit( cache=True, nogil=True )           
def _numba_funcH( v1, v2, dVar ):
    
    h = v1 - v2
    h = np.linalg.norm( h ) ** 2.0
    h /= 4 * ( dVar ** 2.0 )
    h = np.exp( - h )
    return h




@optional_njit( cache=True, nogil=True )           
def _numba_GetMatrixH_aux( Z, dVar ):
    
    dimZ = Z.shape[ 0 ]
    matrixH = np.zeros( ( dimZ, dimZ ) )
    for k1 in range( dimZ ):
        for k2 in range( dimZ ):
            matrixH[ k1, k2 ] = _numba_funcH( Z[ k1, : ], Z[ k2, : ], dVar )

    return matrixH



def _GetMatrixH( Z, dVar ):
    
    dimZ = Z.shape[ 0 ]
    matrixH = _numba_GetMatrixH_aux( Z, dVar )
    sumH = np.sum( np.ravel( matrixH ) ) / perm( dimZ, 2, exact = True )
    matrixH -= sumH / 2.0

    return matrixH



def _getPhi( i, j, matrixH ):
    
    return matrixH[ i, j ] - _funcG( i, matrixH ) - _funcG( j, matrixH )
    


def _funcG( i, matrixH ):
    
    g = 0.0
    for k in range( np.size( matrixH, 0 ) ):
        if k == i: continue
        g += matrixH[ i, k ]
    g /= perm( np.size( matrixH, 0 ), 2, exact = True ) - 2
    
    return g



def _getQHat( X, Y, dVar ):
    
    sizeX = np.size( X, 0 )
    sizeY = np.size( Y, 0 )
    
    QHat1 = 0.0
    for k1 in range( sizeX ):
        for k2 in range( sizeX ):
            QHat1 += _numba_funcH( X[k1, :], X[k2, :], dVar )
    t = perm( sizeX, 2, exact = True )
    if t == 0: QHat1 = 0.0
    else: QHat1 /= t
    
    QHat2 = 0.0
    for k1 in range( sizeY ):
        for k2 in range( sizeY ):
            QHat2 += _numba_funcH( Y[k1, :], Y[k2, :], dVar )
    t = perm( sizeY, 2, exact = True )
    if t == 0: QHat2 = 0.0
    else: QHat2 /= t
    
    QHat3 = 0.0
    for k1 in range( sizeX ):
        for k2 in range( np.size( Y, 0 ) ):
            QHat3 += _numba_funcH( X[k1, :], Y[k2, :], dVar )
    QHat3 *= 2.0
    t = sizeX * sizeY
    if t == 0: QHat3 = 0.0
    else: QHat3 /= t
    
    QHat = QHat1 + QHat2 - QHat3

    return QHat



def _getVariance( X, Y, dVar ):
    
    Z = np.vstack( ( X, Y ) )
    
    matrixH = _GetMatrixH( Z, dVar )
    
    G = np.zeros( ( np.size( Z, 0 ) ) )
    for k1 in range( np.size( Z, 0 ) ):
        G[ k1 ] = _funcG( k1, matrixH )
    
    variance = 0.0
    for k1 in range( np.size( Z, 0 ) ):
        for k2 in range( np.size( Z, 0 ) ):
            variance += ( matrixH[ k1, k2 ] - G[ k1 ] - G[ k2 ] ) ** 2.0
            
    variance /= perm( np.size( Z, 0 ), 2, exact = True )
    variance *= 2.0 * ( ( np.size( Z, 0 ) - 1 ) ** 2 ) * ( np.size( Z, 0 ) - 2 )
    variance /= np.size( X, 0 ) * ( np.size( X, 0 ) - 1 ) * np.size( Y, 0 ) * ( np.size( Y, 0 ) - 1 ) * ( np.size( Z, 0 ) - 3 )

    return variance




def GetPValue( TS, **kwargs ):

    """
    Get p-value according to the Diks' test.

    Parameters
    ----------
    TS : numpy.array
        Time series to be analysed.
    embD : int
        Embedding dimension. Optional, default: 1.
    dVar : float
        Threshold, relative to the standard deviation of the time series, against which values are defined as positive or negative. Optional, default: 1.5.

    Returns
    -------
    float
        p-value of the test.
    float
        statistic of the test.

    Raises
    ------
    ValueError
        If the parameters are not correct.
    """


    # Setting up the parameters

    TS = np.copy( TS )
    embD = kwargs.get( 'embD', 6 )
    dVar = kwargs.get( 'dVar', 1.5 )


    # Checking the parameters

    if type( embD ) is not int:
        raise ValueError("embD is not an integer")
    if embD <= 0:
        raise ValueError("embD must be larger than 0")
    if type( dVar ) is not float:
        raise ValueError("dVar is not a float")
    if dVar <= 0:
        raise ValueError("dVar must be larger than 0")


    # Calculate test
    
    TS -= np.mean( TS )
    TS /= np.std( TS )

    X = _numba_CreateVectors( TS, embD )
    Y = _numba_CreateVectors( TS[ ::-1 ], embD )

    QHat = _getQHat( X, Y, dVar )
    variance = _getVariance( X, Y, dVar )
    estimatorS = QHat / np.sqrt( variance )  
    pValue = -1
    pValue = st.norm.sf( estimatorS )

    return pValue, QHat




def GetStatistic( TS, **kwargs ):

    """
    Get the statistic according to the Dik's test.

    Parameters
    ----------
    TS : numpy.array
        Time series to be analysed.
    embD : int
        Embedding dimension. Optional, default: 1.
    dVar : float
        Threshold, relative to the standard deviation of the time series, against which values are defined as positive or negative. Optional, default: 1.5.

    Returns
    -------
    float
        statistic of the test.

    Raises
    ------
    ValueError
        If the parameters are not correct.
    """

    return GetPValue( TS, numRndReps = 0, **kwargs )[ 1 ]





"""
Trend Pattern Lengths test

From:
    Morales Herrera, J., & Salgado-García, R. (2024).
    Measuring irreversibility via trend pattern lengths.
    AIP Advances, 14(3).
"""

import numpy as np
from ..optionalNJIT import optional_njit
from scipy.stats import ks_2samp



def _mySign( x ):

    return int( x > 0 )




def GetPValue( TS, **kwargs ):

    """
    Get p-value according to the Trend Pattern Lengths test.

    Parameters
    ----------
    TS : numpy.array
        Time series to be analysed.

    Returns
    -------
    float
        p-value of the test.
    float
        statistic of the test.

    Raises
    ------
    ValueError
        If the parameters are not correct.
    """


    dTS = TS[ 1: ] - TS[ :-1 ]

    sign = _mySign( TS[ 0 ] )
    segmL_U = []
    segmL_D = []
    start = 0

    for k in range( 1, np.size( dTS ) ):
        if _mySign( dTS[ k ] ) != sign:

            if sign == 0:
                segmL_D.append( k - start )
            else:
                segmL_U.append( k - start )

            sign = _mySign( dTS[ k ] )
            start = k
            continue

    if sign == 0:
        segmL_D.append( np.size( dTS ) - start )
    else:
        segmL_U.append( np.size( dTS ) - start )

    maxL = np.max( segmL_D + segmL_U )
    prob_U = np.zeros( ( maxL ) )
    for k in segmL_U:
        prob_U[ k - 1 ] += 1

    prob_D = np.zeros( ( maxL ) )
    for k in segmL_D:
        prob_D[ k - 1 ] += 1

    totValues = prob_U + prob_D
    prob_U = prob_U[ totValues > 0 ]
    prob_D = prob_D[ totValues > 0 ]

    if np.sum( prob_U ) == 1 or np.sum( prob_D ) == 1:
        return 0.0, 0.0
    
    from scipy.stats import chi2_contingency
    res = chi2_contingency( [ prob_U, prob_D ] )


    from scipy.stats import entropy
    e = entropy( prob_U, prob_D )

    return res[ 1 ], e




def GetStatistic( TS, **kwargs ):

    """
    Get the statistic according to the Trend Pattern Lengths test.

    Parameters
    ----------
    TS : numpy.array
        Time series to be analysed.

    Returns
    -------
    float
        statistic of the test.

    Raises
    ------
    ValueError
        If the parameters are not correct.
    """


    return GetPValue( TS, **kwargs )[ 1 ]

